from __future__ import annotations

import warnings
from dataclasses import asdict, dataclass
from typing import Tuple

import numpy as np
import torch

from hivemind.proto.runtime_pb2 import CompressionType
from hivemind.utils.serializer import MSGPackSerializer

DUMMY_BATCH_SIZE = 3  # used for dummy runs only

warnings.filterwarnings("ignore", "CUDA initialization*", category=UserWarning)


# ^-- cures https://github.com/pytorch/pytorch/issues/47038


@dataclass(init=True, repr=True, frozen=True)
class DescriptorBase:
    pass


@dataclass(init=True, repr=True, frozen=True)
class TensorDescriptor(DescriptorBase):
    size: tuple
    dtype: torch.dtype = None
    layout: torch.layout = torch.strided
    device: torch.device = None
    requires_grad: bool = False
    pin_memory: bool = False
    compression: CompressionType = CompressionType.NONE

    @property
    def shape(self) -> Tuple[int, ...]:
        return self.size

    def numel(self) -> int:
        return int(np.prod(self.size))

    @classmethod
    def from_tensor(cls, tensor: torch.Tensor) -> TensorDescriptor:
        return cls(
            tensor.shape, tensor.dtype, tensor.layout, tensor.device, tensor.requires_grad, _safe_check_pinned(tensor)
        )

    def make_zeros(self, **kwargs):
        properties = asdict(self)
        properties.update(kwargs)
        properties.pop("compression")
        return torch.zeros(**properties)


def _str_to_torch_type(name: str, torch_type: type):
    try:
        value = getattr(torch, name.split(".")[-1])
    except AttributeError:
        raise ValueError(f"Invalid dtype: torch has no attribute {name}")
    if not isinstance(value, torch_type):
        raise ValueError(f"Invalid dtype: expected {torch_type}, got: {type(value)}")

    return value


@MSGPackSerializer.ext_serializable(0x51)
@dataclass(repr=True, frozen=True)
class BatchTensorDescriptor(TensorDescriptor):
    """torch.Tensor with a variable 0-th dimension, used to describe batched data"""

    def __init__(self, *instance_size, **kwargs):  # compatibility: allow initializing with *size
        if len(instance_size) == 1 and isinstance(instance_size[0], (list, tuple, torch.Size)):
            instance_size = instance_size[0]  # we were given size as the only parameter instead of *parameters
        super().__init__((None, *instance_size), **kwargs)

    @classmethod
    def from_tensor(cls, tensor: torch.Tensor, compression=CompressionType.NONE) -> BatchTensorDescriptor:
        return cls(
            *tensor.shape[1:],
            dtype=tensor.dtype,
            layout=tensor.layout,
            device=tensor.device,
            requires_grad=tensor.requires_grad,
            pin_memory=_safe_check_pinned(tensor),
            compression=compression if tensor.is_floating_point() else CompressionType.NONE,
        )

    def make_zeros(self, *batch_size: int, **kwargs) -> torch.Tensor:
        assert self.shape[0] is None, "Make sure 0-th dimension is not specified (set to None)"
        return super().make_zeros(size=(*batch_size, *self.shape[1:]), **kwargs)

    def packb(self) -> bytes:
        obj_dict = asdict(self)

        obj_dict["dtype"] = str(self.dtype) if self.dtype is not None else None
        obj_dict["layout"] = str(self.layout) if self.layout is not None else None

        device = obj_dict.pop("device")
        device_type, device_index = (device.type, device.index) if device is not None else (None, None)
        obj_dict.update(
            device_type=device_type,
            device_index=device_index,
        )

        return MSGPackSerializer.dumps(obj_dict)

    @classmethod
    def unpackb(cls, raw: bytes) -> BatchTensorDescriptor:
        obj_dict = MSGPackSerializer.loads(raw)

        if obj_dict["dtype"] is not None:
            obj_dict["dtype"] = _str_to_torch_type(obj_dict["dtype"], torch.dtype)

        if obj_dict["layout"] is not None:
            obj_dict["layout"] = _str_to_torch_type(obj_dict["layout"], torch.layout)

        if obj_dict["device_type"] is not None:
            obj_dict["device"] = torch.device(obj_dict["device_type"], obj_dict["device_index"])
        else:
            obj_dict["device"] = None

        del obj_dict["device_type"], obj_dict["device_index"]

        size = obj_dict.pop("size")[1:]

        return BatchTensorDescriptor(*size, **obj_dict)


def _safe_check_pinned(tensor: torch.Tensor) -> bool:
    """Check whether or not a tensor is pinned. If torch cannot initialize cuda, returns False instead of error."""
    try:
        return torch.cuda.is_available() and tensor.is_pinned()
    except RuntimeError:
        return False



import dataclasses
import warnings
from abc import ABC, abstractmethod
from enum import Enum, auto
from typing import Any, Optional

import numpy as np
import torch

from hivemind.proto import runtime_pb2
from hivemind.utils.tensor_descr import TensorDescriptor

# While converting read-only NumPy arrays into PyTorch tensors, we don't make extra copies for efficiency
warnings.filterwarnings("ignore", message="The given NumPy array is not writable", category=UserWarning)


Key = Any


class TensorRole(Enum):
    ACTIVATION = auto()
    PARAMETER = auto()
    GRADIENT = auto()
    OPTIMIZER = auto()
    UNSPECIFIED = auto()


@dataclasses.dataclass(frozen=True)
class CompressionInfo:
    """Auxiliary data structure that contains information about the tensor that determines how it is compressed"""

    key: Key  # name or index of the tensor from named parameters, optimizer state dict or i/o structure
    descriptor: TensorDescriptor  # data structure that defines shape, dtype, layout and device information
    role: TensorRole = TensorRole.UNSPECIFIED  # which role does the tensor play with respect to the model
    part_index: int = 0  # if tensor is sliced into parts, this represents the index within one tensor
    part_size: Optional[int] = None  # if tensor is sliced into parts, this is the _maximum_ number of values per part

    @classmethod
    def from_tensor(cls, tensor: torch.Tensor, key: Key = None, descriptor: TensorDescriptor = None, **kwargs):
        return cls(key, descriptor or TensorDescriptor.from_tensor(tensor), **kwargs)

    def get_part(self, part_index: int, part_size: Optional[int]):
        return CompressionInfo(self.key, self.descriptor, self.role, part_index=part_index, part_size=part_size)


class CompressionBase(ABC):
    """A base class that applies compression algorithm to a pytorch tensor"""

    compression_type: runtime_pb2.CompressionType

    @abstractmethod
    def compress(self, tensor: torch.Tensor, info: CompressionInfo, allow_inplace: bool = False) -> runtime_pb2.Tensor:
        """
        Applies compression algorithm to a tensor based on their meta-parameters

        :param tensor: a pytorch tensor to compress; depending on the application, it is a full tensor or a part
        :param info: meta-information about the tensor; if partitioning is used, this still describes the full tensor
        :param allow_inplace: if True, compression can (but doesn't have to) to modify tensor in-place for efficiency
        :returns: a protobuf message that encodes the tensor
        """
        ...

    @abstractmethod
    def extract(self, serialized_tensor: runtime_pb2.Tensor) -> torch.Tensor:
        """Create a pytorch tensor from the serialized outputs of .compress"""
        ...

    @abstractmethod
    def estimate_compression_ratio(self, info: CompressionInfo) -> float:
        """Estimate the compression ratio without doing the actual compression; lower ratio = better compression"""
        ...

    def __repr__(self):
        return f"hivemind.{self.__class__.__name__}()"


class NoCompression(CompressionBase):
    """A dummy compression strategy that preserves the original tensor as is."""

    compression_type = runtime_pb2.CompressionType.NONE

    def compress(self, tensor: torch.Tensor, info: CompressionInfo, allow_inplace: bool = False) -> runtime_pb2.Tensor:
        tensor = tensor.detach()
        dtype_name = str(tensor.dtype).lstrip("torch.")
        if tensor.dtype == torch.bfloat16:
            tensor = tensor.to(torch.float32)

        return runtime_pb2.Tensor(
            compression=self.compression_type,
            buffer=tensor.numpy().tobytes(),
            size=tensor.shape,
            dtype=dtype_name,
            requires_grad=tensor.requires_grad,
        )

    def extract(self, serialized_tensor: runtime_pb2.Tensor) -> torch.Tensor:
        if serialized_tensor.dtype == "bfloat16":
            array = np.frombuffer(serialized_tensor.buffer, dtype=np.float32)
            tensor = torch.as_tensor(array, dtype=torch.bfloat16)
        else:
            array = np.frombuffer(serialized_tensor.buffer, dtype=np.dtype(serialized_tensor.dtype))
            tensor = torch.as_tensor(array)
        return tensor.reshape(tuple(serialized_tensor.size))

    def estimate_compression_ratio(self, info: CompressionInfo) -> float:
        return 1.0



import os
import threading
from datetime import datetime
from pathlib import Path
from shutil import copy2
from tempfile import TemporaryDirectory
from typing import Dict

import torch

from hivemind.moe.server.module_backend import ModuleBackend
from hivemind.utils.logging import get_logger

logger = get_logger(__name__)


def is_directory(directory: Path):
    assert directory is not None
    assert directory.exists()
    assert directory.is_dir()
    return True


def copy_tree(src: str, dst: str):
    if not os.path.exists(dst):
        os.makedirs(dst)
    for item in os.listdir(src):
        src_entry = os.path.join(src, item)
        dst_entry = os.path.join(dst, item)
        if os.path.isdir(src_entry):
            copy_tree(src_entry, dst_entry)
        else:
            copy2(src_entry, dst_entry)


class CheckpointSaver(threading.Thread):
    def __init__(self, module_backends: Dict[str, ModuleBackend], checkpoint_dir: Path, update_period: float):
        super().__init__()
        assert is_directory(checkpoint_dir)
        self.module_backends = module_backends
        self.update_period = update_period
        self.checkpoint_dir = checkpoint_dir
        self.stop = threading.Event()

        # create expert directories to ensure that the directory is writable and checkpoints can be loaded
        store_experts(self.module_backends, self.checkpoint_dir)

    def run(self) -> None:
        while not self.stop.wait(self.update_period):
            store_experts(self.module_backends, self.checkpoint_dir)


def store_experts(experts: Dict[str, ModuleBackend], checkpoint_dir: Path):
    logger.debug(f"Storing experts at {checkpoint_dir.absolute()}")
    assert is_directory(checkpoint_dir)
    timestamp = datetime.now().isoformat(sep="_")
    with TemporaryDirectory() as tmpdirname:
        for expert_name, expert_backend in experts.items():
            expert_dir = Path(tmpdirname) / expert_name
            expert_dir.mkdir()
            checkpoint_name = expert_dir / f"checkpoint_{timestamp}.pt"
            torch.save(expert_backend.state_dict(), checkpoint_name)
            os.symlink(checkpoint_name, expert_dir / "checkpoint_last.pt")
        copy_tree(tmpdirname, str(checkpoint_dir))


def load_experts(experts: Dict[str, ModuleBackend], checkpoint_dir: Path):
    assert is_directory(checkpoint_dir)
    for expert_name, expert in experts.items():
        checkpoints_folder = checkpoint_dir / expert_name
        latest_checkpoint = checkpoints_folder / "checkpoint_last.pt"
        if latest_checkpoint.exists():
            expert.load_state_dict(torch.load(latest_checkpoint))
        else:
            logger.warning(f"Failed to load checkpoint for expert {expert_name}")



import logging
import os
import sys
import threading
from enum import Enum
from typing import Any, Optional, Union


def in_ipython() -> bool:
    """Check if the code is run in IPython, Jupyter, or Colab"""

    try:
        __IPYTHON__
        return True
    except NameError:
        return False


logging.addLevelName(logging.WARNING, "WARN")
loglevel = os.getenv("HIVEMIND_LOGLEVEL", "INFO")

TRUE_CONSTANTS = ["TRUE", "1"]

_env_colors = os.getenv("HIVEMIND_COLORS")
if _env_colors is not None:
    use_colors = _env_colors.upper() in TRUE_CONSTANTS
else:
    use_colors = sys.stderr.isatty() or in_ipython()

_env_log_caller = os.getenv("HIVEMIND_ALWAYS_LOG_CALLER", "0")
always_log_caller = _env_log_caller.upper() in TRUE_CONSTANTS


class HandlerMode(Enum):
    NOWHERE = 0
    IN_HIVEMIND = 1
    IN_ROOT_LOGGER = 2


_init_lock = threading.RLock()
_current_mode = HandlerMode.IN_HIVEMIND
_default_handler = None


class _DisableIfNoColors(type):
    def __getattribute__(self, name: str) -> Any:
        if name.isupper() and not use_colors:
            return ""
        return super().__getattribute__(name)


class TextStyle(metaclass=_DisableIfNoColors):
    """
    ANSI escape codes. Details: https://en.wikipedia.org/wiki/ANSI_escape_code#Colors
    """

    RESET = "\033[0m"
    BOLD = "\033[1m"
    RED = "\033[31m"
    BLUE = "\033[34m"
    PURPLE = "\033[35m"
    ORANGE = "\033[38;5;208m"  # From 8-bit palette


class CustomFormatter(logging.Formatter):
    """
    A formatter that allows a log time and caller info to be overridden via
    ``logger.log(level, message, extra={"origin_created": ..., "caller": ...})``.
    """

    # Details: https://en.wikipedia.org/wiki/ANSI_escape_code#Colors
    _LEVEL_TO_COLOR = {
        logging.DEBUG: TextStyle.PURPLE,
        logging.INFO: TextStyle.BLUE,
        logging.WARNING: TextStyle.ORANGE,
        logging.ERROR: TextStyle.RED,
        logging.CRITICAL: TextStyle.RED,
    }

    def format(self, record: logging.LogRecord) -> str:
        if hasattr(record, "origin_created"):
            record.created = record.origin_created
            record.msecs = (record.created - int(record.created)) * 1000

        if record.levelno != logging.INFO or always_log_caller:
            if not hasattr(record, "caller"):
                record.caller = f"{record.name}.{record.funcName}:{record.lineno}"
            record.caller_block = f" [{TextStyle.BOLD}{record.caller}{TextStyle.RESET}]"
        else:
            record.caller_block = ""

        # Aliases for the format argument
        record.levelcolor = self._LEVEL_TO_COLOR[record.levelno]
        record.bold = TextStyle.BOLD
        record.reset = TextStyle.RESET

        return super().format(record)


def _initialize_if_necessary():
    global _current_mode, _default_handler

    with _init_lock:
        if _default_handler is not None:
            return

        formatter = CustomFormatter(
            fmt="{asctime}.{msecs:03.0f} [{bold}{levelcolor}{levelname}{reset}]{caller_block} {message}",
            style="{",
            datefmt="%b %d %H:%M:%S",
        )
        _default_handler = logging.StreamHandler()
        _default_handler.setFormatter(formatter)

        _enable_default_handler("hivemind")


def get_logger(name: Optional[str] = None) -> logging.Logger:
    """
    Same as ``logging.getLogger()`` but ensures that the default hivemind log handler is initialized.

    :note: By default, the hivemind log handler (that reads the ``HIVEMIND_LOGLEVEL`` env variable and uses
           the colored log formatter) is only applied to messages logged inside the hivemind package.
           If you want to extend this handler to other loggers in your application, call
           ``use_hivemind_log_handler("in_root_logger")``.
    """

    _initialize_if_necessary()
    return logging.getLogger(name)


def _enable_default_handler(name: Optional[str]) -> None:
    logger = get_logger(name)

    # Remove the extra default handler in the Colab's default logger before adding a new one
    if isinstance(logger, logging.RootLogger):
        for handler in list(logger.handlers):
            if isinstance(handler, logging.StreamHandler) and handler.stream is sys.stderr:
                logger.removeHandler(handler)

    logger.addHandler(_default_handler)
    logger.propagate = False
    logger.setLevel(loglevel)


def _disable_default_handler(name: Optional[str]) -> None:
    logger = get_logger(name)
    logger.removeHandler(_default_handler)
    logger.propagate = True
    logger.setLevel(logging.NOTSET)


def use_hivemind_log_handler(where: Union[HandlerMode, str]) -> None:
    """
    Choose loggers where the default hivemind log handler is applied. Options for the ``where`` argument are:

    * "in_hivemind" (default): Use the hivemind log handler in the loggers of the ``hivemind`` package.
                               Don't propagate their messages to the root logger.
    * "nowhere": Don't use the hivemind log handler anywhere.
                 Propagate the ``hivemind`` messages to the root logger.
    * "in_root_logger": Use the hivemind log handler in the root logger
                        (that is, in all application loggers until they disable propagation to the root logger).
                        Propagate the ``hivemind`` messages to the root logger.

    The options may be defined as strings (case-insensitive) or values from the HandlerMode enum.
    """

    global _current_mode

    if isinstance(where, str):
        # We allow `where` to be a string, so a developer does not have to import the enum for one usage
        where = HandlerMode[where.upper()]

    _initialize_if_necessary()

    if where == _current_mode:
        return

    if _current_mode == HandlerMode.IN_HIVEMIND:
        _disable_default_handler("hivemind")
    elif _current_mode == HandlerMode.IN_ROOT_LOGGER:
        _disable_default_handler(None)

    _current_mode = where

    if _current_mode == HandlerMode.IN_HIVEMIND:
        _enable_default_handler("hivemind")
    elif _current_mode == HandlerMode.IN_ROOT_LOGGER:
        _enable_default_handler(None)


def golog_level_to_python(level: str) -> int:
    level = level.upper()
    if level in ["DPANIC", "PANIC", "FATAL"]:
        return logging.CRITICAL

    level = logging.getLevelName(level)
    if not isinstance(level, int):
        raise ValueError(f"Unknown go-log level: {level}")
    return level


def python_level_to_golog(level: str) -> str:
    if not isinstance(level, str):
        raise ValueError("`level` is expected to be a Python log level in the string form")

    if level == "CRITICAL":
        return "FATAL"
    if level == "WARNING":
        return "WARN"
    return level



from ipaddress import ip_address
from typing import List, Sequence

from multiaddr import Multiaddr

from hivemind.utils.logging import TextStyle, get_logger

LOCALHOST = "127.0.0.1"

logger = get_logger(__name__)


def choose_ip_address(
    maddrs: Sequence[Multiaddr], prefer_global: bool = True, protocol_priority: Sequence[str] = ("ip4", "ip6")
) -> str:
    """
    Currently, some components of hivemind are not converted to work over libp2p and use classical networking.
    To allow other peers reach a server when needed, these components announce a machine's IP address.

    This function automatically selects the best IP address to announce among publicly visible multiaddrs
    of this machine identified by libp2p (typically, using the ``P2P.get_visible_maddrs()`` method),
    so a user does not need to define this address manually (unless the user wants to).

    The best IP address is chosen using the following logic:
      - Prefer IP addresses from global address blocks
        (in terms of https://docs.python.org/3/library/ipaddress.html#ipaddress.IPv4Address.is_global)
      - Among the IP addresses of the same globality status, prefer IPv4 addresses over IPv6

    If the default logic does not suit you, it is recommended to set the announced IP address manually.
    """

    for need_global in [prefer_global, not prefer_global]:
        for protocol in protocol_priority:
            for addr in maddrs:
                if protocol in addr.protocols():
                    value_for_protocol = addr[protocol]
                    if ip_address(value_for_protocol).is_global == need_global:
                        return value_for_protocol

    raise ValueError(f"No IP address found among given multiaddrs: {maddrs}")


def log_visible_maddrs(visible_maddrs: List[Multiaddr], only_p2p: bool) -> None:
    if only_p2p:
        unique_addrs = {addr["p2p"] for addr in visible_maddrs}
        initial_peers = " ".join(f"/p2p/{addr}" for addr in unique_addrs)
    else:
        available_ips = [Multiaddr(addr) for addr in visible_maddrs if "ip4" in addr or "ip6" in addr]
        if available_ips:
            preferred_ip = choose_ip_address(available_ips)
            selected_maddrs = [addr for addr in visible_maddrs if preferred_ip in str(addr)]
        else:
            selected_maddrs = visible_maddrs
        initial_peers = " ".join(str(addr) for addr in selected_maddrs)

    logger.info(
        f"Running a DHT instance. To connect other peers to this one, use "
        f"{TextStyle.BOLD}{TextStyle.BLUE}--initial_peers {initial_peers}{TextStyle.RESET}"
    )
    logger.info(f"Full list of visible multiaddresses: {' '.join(str(addr) for addr in visible_maddrs)}")



import os
import struct
from enum import Enum
from typing import Optional

import numpy as np
import torch

from hivemind.utils import DHTExpiration, MPFuture, get_dht_time, get_logger

logger = get_logger(__name__)


class AveragingStage(Enum):
    IDLE = 0  # still initializing
    LOOKING_FOR_GROUP = 1  # running decentralized matchmaking, can't run allreduce yet
    AWAITING_TRIGGER = 2  # waiting for user to set the trigger that allows running allreduce
    RUNNING_ALLREDUCE = 3  # exchanging tensors with groupmates
    FINISHED = 4  # either done or failed with exception


class StepControl(MPFuture):
    """
    An auxiliary data structure that allows user to control stages and track progress in a single averaging step

    :param scheduled_time: estimated time when averaging should begin. Will be used for scheduling
    :param deadline: if averaging is still in progress at this time, it should be stopped due to TimeoutError
    :param allow_retries: if True, allow running matchmaking and all-reduce again if previous attempt fails
    :param weight: averaging weight, can be changed afterwards
    :param data_for_gather: send this data to all peers in the next group and gather it from groupmates
    """

    # indices for the shared buffer
    _SCHEDULED_TIME, _WEIGHT, _STAGE, _BEGAN_ALLREDUCE = slice(0, 8), slice(8, 16), 16, 17

    def __init__(
        self,
        scheduled_time: DHTExpiration,
        deadline: float,
        allow_retries: bool,
        weight: float,
        data_for_gather: bytes,
    ):
        super().__init__()
        self._data_for_gather, self._deadline, self._allow_retries = data_for_gather, deadline, allow_retries
        self._trigger: Optional[MPFuture] = None
        self._cancel: Optional[MPFuture] = None

        # Buffer contents:
        # scheduled_time (double) | weight (double) | stage (AveragingStage, 1 byte) | began_allreduce: (bool, 1 byte)
        self._shared_buffer = torch.zeros([18], dtype=torch.uint8).share_memory_()
        self.stage = AveragingStage.IDLE
        self.scheduled_time = scheduled_time
        self.weight = weight
        self.began_allreduce = False

    def attach(self, trigger: MPFuture, cancel: MPFuture):
        assert self._trigger is None and self._cancel is None, "Futures are already attached"
        self._trigger, self._cancel = trigger, cancel

    def allow_allreduce(self):
        """Allow averager to begin all-reduce when it finds a group. Meant to be triggered by user."""
        assert self._trigger is not None, "StepControl does not have an attached trigger"
        if self._trigger.done():
            logger.warning("Trigger is already set")
        else:
            self._trigger.set_result(None)

    async def wait_for_trigger(self):
        assert self._trigger is not None, "StepControl does not have an attached trigger"
        await self._trigger

    @property
    def triggered(self) -> bool:
        assert self._trigger is not None, "StepControl does not have an attached trigger"
        return self._trigger.done()

    @property
    def scheduled_time(self) -> DHTExpiration:
        return struct.unpack("d", self._shared_buffer[StepControl._SCHEDULED_TIME].numpy().data)[0]

    @scheduled_time.setter
    def scheduled_time(self, scheduled_time):
        if self.began_allreduce:
            logger.warning("Changing scheduled time has no effect after all-reduce has already started")
        if scheduled_time >= self.deadline:
            logger.warning("Changing scheduled time to after deadline, averaging will likely fail due to timeout")
        struct.pack_into("d", self._shared_buffer[StepControl._SCHEDULED_TIME].numpy().data, 0, float(scheduled_time))

    @property
    def weight(self) -> float:
        return struct.unpack("d", self._shared_buffer[StepControl._WEIGHT].numpy().data)[0]

    @weight.setter
    def weight(self, weight: float):
        assert weight >= 0 and np.isfinite(weight)
        if self.began_allreduce:
            logger.warning("Changing weights has no effect after all-reduce has already started")
        struct.pack_into("d", self._shared_buffer[StepControl._WEIGHT].numpy().data, 0, float(weight))

    @property
    def stage(self) -> AveragingStage:
        return AveragingStage(self._shared_buffer[StepControl._STAGE].item())

    @stage.setter
    def stage(self, stage: AveragingStage):
        if stage == AveragingStage.RUNNING_ALLREDUCE:
            self.began_allreduce = True
        self._shared_buffer[StepControl._STAGE] = stage.value

    @property
    def began_allreduce(self) -> bool:
        return bool(self._shared_buffer[StepControl._BEGAN_ALLREDUCE].item())

    @began_allreduce.setter
    def began_allreduce(self, value: bool):
        self._shared_buffer[StepControl._BEGAN_ALLREDUCE] = int(value)

    @property
    def data_for_gather(self) -> bytes:
        return self._data_for_gather

    @property
    def deadline(self) -> DHTExpiration:
        return self._deadline

    def get_timeout(self) -> Optional[DHTExpiration]:
        return max(0.0, self.deadline - get_dht_time())

    @property
    def allow_retries(self) -> bool:
        return self._allow_retries

    def __getstate__(self):
        return dict(
            super().__getstate__(),
            _trigger=self._trigger,
            _cancel=self._cancel,
            _shared_buffer=self._shared_buffer,
            immutable_params=(self._data_for_gather, self._deadline, self._allow_retries),
        )

    def __setstate__(self, state):
        super().__setstate__(state)
        self._trigger, self._cancel, self._shared_buffer = state["_trigger"], state["_cancel"], state["_shared_buffer"]
        self._data_for_gather, self._deadline, self._allow_retries = state["immutable_params"]

    def __del__(self):
        if os.getpid() == self._origin_pid and not self.triggered:
            logger.warning(
                "Deleted an averaging StepControl, but the step was not triggered. This may cause other "
                "peers to fail an averaging round via TimeoutError."
            )
        super().__del__()

    def cancel(self) -> bool:
        if self._trigger is not None:
            self._trigger.cancel()
        if self._cancel is not None:
            self._cancel.set_result(None)
        return super().cancel()

    async def wait_for_cancel(self):
        """Await for step to be cancelled by the user. Should be called from insider the averager."""
        await self._cancel



"""
Utilities for streaming tensors
"""

from __future__ import annotations

from typing import Iterable, Iterator

from hivemind.proto import runtime_pb2
from hivemind.utils.logging import get_logger

logger = get_logger(__name__)

STREAMING_CHUNK_SIZE_BYTES = 2**16


def split_for_streaming(
    serialized_tensor: runtime_pb2.Tensor,
    chunk_size_bytes: int = STREAMING_CHUNK_SIZE_BYTES,
) -> Iterator[runtime_pb2.Tensor]:
    """Split serialized_tensor into multiple chunks for streaming"""
    buffer = memoryview(serialized_tensor.buffer)
    num_chunks = len(range(0, len(buffer), chunk_size_bytes))
    yield runtime_pb2.Tensor(
        compression=serialized_tensor.compression,
        buffer=buffer[:chunk_size_bytes].tobytes(),
        chunks=num_chunks,
        size=serialized_tensor.size,
        dtype=serialized_tensor.dtype,
        requires_grad=serialized_tensor.requires_grad,
    )
    for chunk_start in range(chunk_size_bytes, len(buffer), chunk_size_bytes):
        yield runtime_pb2.Tensor(buffer=buffer[chunk_start : chunk_start + chunk_size_bytes].tobytes())


def combine_from_streaming(stream: Iterable[runtime_pb2.Tensor]) -> runtime_pb2.Tensor:
    """Restore a result of split_into_chunks into a single serialized tensor"""
    stream = iter(stream)
    first_chunk = next(stream)
    serialized_tensor = runtime_pb2.Tensor()
    serialized_tensor.CopyFrom(first_chunk)
    buffer_chunks = [first_chunk.buffer]
    for tensor_part in stream:
        buffer_chunks.append(tensor_part.buffer)
    serialized_tensor.buffer = b"".join(buffer_chunks)
    return serialized_tensor



import time

import torch
from torch import nn as nn

from hivemind.moe.server.layers.custom_experts import register_expert_class


# https://github.com/huggingface/transformers/blob/master/src/transformers/activations.py
@torch.jit.script
def gelu_fast(x):
    return 0.5 * x * (1.0 + torch.tanh(x * 0.7978845608 * (1.0 + 0.044715 * x * x)))


ffn_sample_input = lambda batch_size, hid_dim: torch.empty((batch_size, hid_dim))


@register_expert_class("ffn", ffn_sample_input)
class FeedforwardBlock(nn.Module):
    def __init__(self, hid_dim):
        super().__init__()
        self.ffn = nn.Linear(hid_dim, 4 * hid_dim)
        self.ffn_output = nn.Linear(4 * hid_dim, hid_dim)
        self.layer_norm = nn.LayerNorm(hid_dim, eps=1e-12)

    def forward(self, x):
        ffn_output = self.ffn(x)
        ffn_output = gelu_fast(ffn_output)
        ffn_output = self.ffn_output(ffn_output)
        return self.layer_norm(x + ffn_output)


class TransformerEncoderLayer(nn.Module):
    """
    A slight modification of torch.nn.TransformerEncoderLayer which allows for torch.jit scripting
    """

    def __init__(self, d_model, nhead, dim_feedforward=2048, dropout=0.1):
        super().__init__()
        self.self_attn = nn.MultiheadAttention(d_model, nhead, dropout=dropout)
        # Implementation of Feedforward model
        self.linear1 = nn.Linear(d_model, dim_feedforward)
        self.dropout = nn.Dropout(dropout)
        self.linear2 = nn.Linear(dim_feedforward, d_model)

        self.norm1 = nn.LayerNorm(d_model)
        self.norm2 = nn.LayerNorm(d_model)
        self.dropout1 = nn.Dropout(dropout)
        self.dropout2 = nn.Dropout(dropout)

        self.activation = gelu_fast

    def forward(self, src, src_key_padding_mask=None):
        # (N, S, E) -> (S, N, E)
        src = src.transpose(0, 1)

        src2 = self.self_attn(src, src, src, key_padding_mask=src_key_padding_mask)[0]
        src = src + self.dropout1(src2)
        src = self.norm1(src)
        src2 = self.linear2(self.dropout(self.activation(self.linear1(src))))
        src = src + self.dropout2(src2)
        src = self.norm2(src)

        # (S, N, E) -> (N, S, E)
        src = src.transpose(0, 1)
        return src


transformer_sample_input = lambda batch_size, hid_dim: (
    torch.empty((batch_size, 128, hid_dim)),
    torch.empty((batch_size, 128), dtype=torch.bool),
)


@register_expert_class("transformer", transformer_sample_input)
class TunedTransformer(TransformerEncoderLayer):
    def __init__(self, hid_dim):
        super().__init__(hid_dim, dim_feedforward=4 * hid_dim, nhead=16)


nop_sample_input = lambda batch_size, hid_dim: torch.empty((batch_size, hid_dim))


@register_expert_class("nop", nop_sample_input)
class NopExpert(nn.Sequential):
    def __init__(self, hid_dim):
        super().__init__()
        self.w = nn.Parameter(torch.zeros(0), requires_grad=True)

    def forward(self, x):
        return x.clone()


@register_expert_class("nop_delay", nop_sample_input)
class DelayedNopExpert(nn.Sequential):
    def __init__(self, hid_dim, delay=0.5):
        super().__init__()
        self.w = nn.Parameter(torch.zeros(0), requires_grad=True)
        self.delay = delay

    def forward(self, x):
        time.sleep(self.delay)
        return x.clone()



import asyncio
import contextlib
from enum import Enum
from typing import Any, Iterable, Optional, Sequence

import torch

from hivemind.averaging.allreduce import AveragingMode
from hivemind.averaging.group_info import GroupInfo
from hivemind.averaging.load_balancing import load_balance_peers
from hivemind.averaging.matchmaking import MatchmakingException
from hivemind.compression import CompressionInfo, TensorRole
from hivemind.dht import DHT
from hivemind.optim.grad_averager import GradientAverager
from hivemind.utils import get_logger
from hivemind.utils.asyncio import enter_asynchronously
from hivemind.utils.math import get_flatten_greedy_dims, orthogonalize_

GatheredData = Any
logger = get_logger(__name__)


class AllReducePhases(Enum):
    PHASE_P = 1
    PHASE_Q = 2


class PowerSGDGradientAverager(GradientAverager):
    """
    A gradient averager that implements PowerSGD compression: https://arxiv.org/abs/1905.13727
    For basic properties and guaranties of gradient averagers, please refer to the base class docstring.
    Put simply, this method approximates large gradient tensors (m,n) with a product of two
    smaller matrices (m,r) by (r,n), where r is a parameter chosen by the user (see averager_rank).

    As a result, PowerSGD only needs to aggregate O((m + n) * r) tensors instead of O(m * n).
    High r, e.g. sqrt(max(m, n)) typically reduce communication by 2-8x without affecting convergence.
    Low r, e.g. 1-8, further accelerate communication, but may converge worse depending on the task.

    To maintain convergence with low r, this averager uses the error feedback strategy. Put simply,
    if some part of the gradient is "lost in compression", it will be added to the next iteration.
    This has two implications: (a) it needs more RAM in order to store the "feedback buffers"
    and (b) if devices stay alive only for one step, training with small rank may converge slower.
    This is because error feedback takes multiple steps to kick in.

    Since not all gradients are matrices, PowerSGD views 3d+ tensors via tensor.flatten(1, -1).
    If a tensor has less than 2 dimensions or does not compress efficiently, it will be aggregated
    normally, i.e. without powerSGD. See min_compression_ratio for details.

    :note: due to the above rule, PowerSGD is *not* shape-invariant. For instance, a
     matrix of shape (256, 256) be compressed differently if you .reshape it to (32, 32, 32).

    :param parameters: pytorch parameters for which to aggregate gradients
    :param averager_rank: rank of compressed gradients
    :param dht: a DHT instance connected to the rest of the swarm. See hivemind.DHT docs
    :param prefix: a unique DHT key used for matchmaking. E.g. this can be your experiment name with optional suffixes
    :param reuse_grad_buffers: if True, use model's .grad buffers for accumulating gradients over multiple steps.
      This is more memory efficient, but it requires that the user does *not* call zero_grad or clip_by_whatever at all
    :param accumulate_grads_on: if specified, accumulate gradients on this device. By default, this will use the same
      device as model parameters. One can specify a different device (e.g. 'cpu' vs 'cuda') to save device memory at
      the cost of extra time per step. If reuse_grad_buffers is True, this parameter has no effect.
    :param client_mode: if False, this averager will accept incoming requests from other peers.
      if True, the averager will only join existing groups where at least one peer has client_mode=False.
      By default, this flag is copied from DHTNode inside the ``dht`` instance.
    :param warn: if True, warn when the averager did not reset accumulators after use or did not use averaging results
    :param min_compression_ratio: apply PowerSGD to a tensor only if it reduces communication by at least this factor,
      otherwise aggregate tensors as is
    :param averaged_grads: if provided, it will be used as a set of averagable gradients
    """

    def __init__(
        self,
        parameters: Iterable[torch.nn.Parameter],
        averager_rank: int,
        *,
        dht: DHT,
        prefix: str,
        reuse_grad_buffers: bool = False,
        accumulate_grads_on: Optional[torch.device] = None,
        client_mode: bool = None,
        warn: bool = True,
        min_compression_ratio: float = 0.5,
        averaged_grads: Optional[Sequence[torch.Tensor]] = None,
        **kwargs,
    ):
        self.rank = averager_rank
        self.parameters = tuple(parameters)
        self._uncompressed_gradients_indexes = set(
            i
            for i, grad in enumerate(self._grads_from_parameters())
            if grad.ndim <= 1
            or (1 - self.rank * sum(get_flatten_greedy_dims(grad)) / grad.numel()) < min_compression_ratio
            # compute how much parameters are left after factorization
        )
        self._ms = [
            torch.zeros_like(grad, device="cpu").share_memory_()
            for idx, grad in enumerate(self._grads_from_parameters())
            if idx not in self._uncompressed_gradients_indexes
        ]
        self._qs = [
            torch.rand((get_flatten_greedy_dims(grad)[1], self.rank), device="cpu").share_memory_()
            for idx, grad in enumerate(self._grads_from_parameters())
            if idx not in self._uncompressed_gradients_indexes
        ]

        super().__init__(
            self.parameters,
            dht=dht,
            prefix=prefix,
            reuse_grad_buffers=reuse_grad_buffers,
            accumulate_grads_on=accumulate_grads_on,
            client_mode=client_mode,
            warn=warn,
            averaged_grads=averaged_grads,
            **kwargs,
        )

    @contextlib.contextmanager
    def _register_allreduce_group(self, group_info: GroupInfo):
        """Register a given group for one or more all-reduce rounds"""
        try:
            for phase in list(AllReducePhases):
                self._running_groups[group_info.group_id + phase.name.encode()] = asyncio.Future()
            self._pending_groups_registered.set()
            yield
        finally:
            for phase in list(AllReducePhases):
                maybe_future = self._running_groups.pop(group_info.group_id + phase.name.encode(), None)
                if maybe_future and not maybe_future.done():
                    logger.warning(f"All-reduce group {group_info.group_id + phase.name.encode()} did not finish.")
            self._pending_groups_registered.set()

    async def _aggregate_with_group(self, group_info: GroupInfo, min_vector_size: int, **kwargs) -> GatheredData:
        """Run aggregation in a given group and update tensors in place, return gathered metadata"""
        try:
            bandwidths, mode_ids, user_gathered_bytes = zip(*map(self.serializer.loads, group_info.gathered))
            user_gathered = dict(zip(group_info.peer_ids, map(self.serializer.loads, user_gathered_bytes)))
            modes = tuple(map(AveragingMode, mode_ids))

            download_bandwidths = [
                thr if mode != AveragingMode.CLIENT else 0.0 for thr, mode in zip(bandwidths, modes)
            ]
            peer_fractions = await asyncio.get_event_loop().run_in_executor(
                None, load_balance_peers, self.total_size, download_bandwidths, min_vector_size
            )

            async with enter_asynchronously(self.get_tensors()) as averaged_grads:
                averaged_grads_via_sgd = [
                    grad for idx, grad in enumerate(averaged_grads) if idx not in self._uncompressed_gradients_indexes
                ]
                for grad, m in zip(averaged_grads_via_sgd, self._ms):
                    m.add_(grad.to(m.device))

                ps = [
                    torch.zeros((get_flatten_greedy_dims(grad)[0], self.rank), device="cpu")
                    for idx, grad in enumerate(averaged_grads_via_sgd)
                ]
                for p, q, m in zip(ps, self._qs, self._ms):
                    # we use reshape for all matrixes because PowerSGD works only with 2d tensors
                    torch.matmul(m.reshape(-1, q.size(0)), q, out=p)

                p_group_id = group_info.group_id + AllReducePhases.PHASE_P.name.encode()
                q_groud_id = group_info.group_id + AllReducePhases.PHASE_Q.name.encode()

                await self._run_allreduce_inplace_(ps, group_info, p_group_id, peer_fractions=peer_fractions, **kwargs)

                for p in ps:
                    orthogonalize_(p)

                for p, q, m in zip(ps, self._qs, self._ms):
                    torch.matmul(m.reshape(-1, q.size(0)).t(), p, out=q)

                phase_q_tensors = self._qs + [
                    grad for idx, grad in enumerate(averaged_grads) if idx in self._uncompressed_gradients_indexes
                ]

                await self._run_allreduce_inplace_(
                    phase_q_tensors, group_info, q_groud_id, peer_fractions=peer_fractions, **kwargs
                )

                for p, q, m, grad in zip(ps, self._qs, self._ms, averaged_grads_via_sgd):
                    new_m = torch.matmul(p, q.t()).reshape(m.size())
                    m.sub_(new_m)
                    grad.copy_(new_m)

                return user_gathered
        except BaseException as e:
            logger.exception(e)
            raise MatchmakingException(f"Unable to run All-Reduce: {e}")

    def get_current_state(self):
        """
        Get current gradient averager state and when requested by a newbie peer.
        """
        with torch.no_grad(), self.lock_averaged_tensors:
            grad_averager_buffers = [q for q in self._qs]
            grad_averager_buffers_infos = [
                CompressionInfo.from_tensor(buffer, key=f"buffer_q_{key}", role=TensorRole.GRADIENT)
                for buffer, key in zip(grad_averager_buffers, enumerate(grad_averager_buffers))
            ]

        metadata = dict(group_bits=self.get_group_bits())
        return metadata, grad_averager_buffers, grad_averager_buffers_infos

    def load_state_from_peers(self, **kwargs):
        """
        Attempt to download the latest optimizer state from peers and update gradient averager buffers.
        :returns: whether or the averager succeeded in loading parameters
        """
        loaded_state = super().load_state_from_peers(**kwargs)
        if loaded_state is None:
            return

        metadata, flat_tensors = loaded_state
        logger.info("Starting loading gradient averager buffers from peers")

        if len(flat_tensors) != len(self._qs):
            logger.error("Failed to load state from peer, received invalid parameters, extras or metadata")
            return

        with torch.no_grad(), self.lock_averaged_tensors:
            for local_q, loaded_q in zip(self._qs, flat_tensors):
                local_q.copy_(loaded_q, non_blocking=True)



import torch
import torch.nn.functional as F


@torch.jit.script
def orthogonalize_(matrix, eps: float = 1e-8):
    """Orthogonalize a 2d tensor in-place over the last dimension"""
    n, m = matrix.shape
    for i in range(m):
        col = matrix[:, i]
        F.normalize(col, dim=0, eps=eps, out=col)
        if i + 1 < m:
            rest = matrix[:, i + 1 :]
            rest.addmm_(col[:, None], (col @ rest)[None, :], alpha=-1)


def get_flatten_greedy_dims(tensor: torch.Tensor, max_ndim: int = 2):
    """get dims to flatten tensor up to max_ndim dimensions by merging small axes together"""
    dims = list(tensor.shape)
    while len(dims) > max_ndim:
        squeeze_ix = min(range(len(dims) - 1), key=lambda i: dims[i] * dims[i + 1])
        squeezed_dim = dims.pop(squeeze_ix)
        dims[squeeze_ix] *= squeezed_dim
    return dims



from __future__ import annotations

import time
from concurrent.futures import Future
from queue import Empty, Queue
from typing import Any, Dict, List, Optional, Tuple

import torch
import torch.nn as nn
from torch.autograd.function import once_differentiable

from hivemind.compression import serialize_torch_tensor
from hivemind.dht import DHT
from hivemind.moe.client.beam_search import MoEBeamSearcher
from hivemind.moe.client.expert import DUMMY, RemoteExpert, expert_backward, expert_forward, get_server_stub
from hivemind.moe.client.remote_expert_worker import RemoteExpertWorker
from hivemind.moe.expert_uid import UID_DELIMITER
from hivemind.p2p.p2p_daemon_bindings.control import P2PDaemonError
from hivemind.utils import nested_flatten, nested_map, nested_pack
from hivemind.utils.logging import get_logger

logger = get_logger(__name__)


class RemoteMixtureOfExperts(nn.Module):
    """
    A torch module that performs Mixture-of-Experts inference with a local gating function and multiple remote experts.
    Natively supports pytorch autograd.

    :note: By default, not all experts are guaranteed to perform forward pass. Moreover, not all of those who ran
     forward pass are guaranteed to perform backward pass. In the latter case, gradient will be averaged without
     the missing experts

    :param in_features: common input size for experts and gating function
    :param grid_size: dimensions that form expert uid (see below)
    :param uid_prefix: common prefix for all expert uids (must end with '.')
    :note: expert uid follows the pattern {uid_prefix}.{0...grid_size[0]}.{0...grid_size[1]}...{0...grid_size[-1]}
    :param dht: a DHT instance used to search for best experts
    :param k_best: average this many highest-scoring experts to compute activations
    :param k_min: make sure at least this many experts returned output (i.e. didn't fail)
    :param timeout_after_k_min: wait for this many seconds after k_min experts returned results.
     Any expert that didn't manage to return output after that delay is considered unavailable
    :param detect_anomalies: whether to check input/output tensors for NaN and infinity values
    :param allow_zero_outputs: whether to return zeros if no experts respond on forward pass
    """

    def __init__(
        self,
        *,
        in_features,
        grid_size: Tuple[int, ...],
        dht: DHT,
        uid_prefix: str,
        k_best: int,
        k_min: int = 1,
        forward_timeout: Optional[float] = None,
        timeout_after_k_min: Optional[float] = None,
        backward_k_min: int = 1,
        backward_timeout: Optional[float] = None,
        detect_anomalies: bool = False,
        allow_zero_outputs: bool = False,
        **dht_kwargs,
    ):
        super().__init__()
        self.dht = dht
        self.beam_search = MoEBeamSearcher(dht, uid_prefix, grid_size, **dht_kwargs)
        self.k_best, self.k_min, self.backward_k_min = k_best, k_min, backward_k_min
        self.forward_timeout, self.backward_timeout = forward_timeout, backward_timeout
        self.timeout_after_k_min = timeout_after_k_min
        self.detect_anomalies = detect_anomalies
        self.allow_zero_outputs = allow_zero_outputs

        # jointly predict logits for all grid dimensions
        self.proj = nn.Linear(in_features, self.beam_search.total_grid_size)
        self._expert_info = None  # expert['info'] from one of experts in the grid

    def forward(self, input: torch.Tensor, *args: torch.Tensor, **kwargs: torch.Tensor):
        """
        Choose k best experts with beam search, then call chosen experts and average their outputs.
        Input tensor is averaged over all dimensions except for first and last
        (we assume that extra dimensions represent sequence length or image height/width)

        :param input: a tensor of values that are used to estimate gating function, batch-first.
        :param args: extra positional parameters that will be passed to each expert after input, batch-first
        :param kwargs: extra keyword parameters that will be passed to each expert, batch-first
        :returns: averaged predictions of all experts that delivered result on time, nested structure of batch-first
        """
        if input.ndim != 2:
            input_for_gating = input.mean(dim=tuple(range(1, input.ndim - 1)))
        else:
            input_for_gating = input

        # 1. compute scores and find most appropriate experts with beam search
        grid_scores = self.proj(input_for_gating).split_with_sizes(self.beam_search.grid_size, dim=-1)

        chosen_experts: List[List[RemoteExpert]] = self.beam_search.batch_find_best_experts(
            [scores.detach().cpu().numpy() for scores in grid_scores], self.k_best
        )

        if self._expert_info is None:
            try:
                self._expert_info = next((expert.info for experts_i in chosen_experts for expert in experts_i))
            except StopIteration:
                raise RuntimeError(
                    "No responding experts found during beam search. Check that UID prefixes and "
                    "the grid size are consistent with running Server instances."
                )
            except P2PDaemonError as e:
                logger.warning(f"Failed to get RemoteMixtureOfExperts.output_shape: {e}")

        expert_mask, *expert_outputs = _RemoteCallMany.apply(
            DUMMY,
            chosen_experts,
            self.k_min,
            self.backward_k_min,
            self.timeout_after_k_min,
            self.forward_timeout,
            self.backward_timeout,
            self.detect_anomalies,
            self.allow_zero_outputs,
            self.info,
            *nested_flatten(((input, *args), kwargs)),
        )
        # ^-- multiple tensors of shape [batch_size, max_experts, ...output_shape]

        expert_logits = self.compute_expert_scores(grid_scores, chosen_experts)
        masked_logits = torch.full((1,), float("-inf"), device=expert_logits.device, dtype=expert_logits.dtype)
        expert_logits = torch.where(expert_mask, expert_logits, masked_logits)
        expert_weights = torch.softmax(expert_logits, dim=1)
        averaged_outputs_flat = [
            (expert_weights[..., None] * tensor.flatten(start_dim=2)).view(tensor.shape).sum(dim=1)
            for tensor in expert_outputs
        ]  # ^-- multiply by softmax weights along first 2 axes

        return nested_pack(averaged_outputs_flat, self.info["outputs_schema"])

    def compute_expert_scores(
        self, grid_scores: List[torch.Tensor], batch_experts: List[List[RemoteExpert]]
    ) -> torch.Tensor:
        """
        Compute scores for each expert by adding up grid scores, autograd-friendly
        :param grid_scores: list of torch tensors, i-th tensor contains scores for i-th grid dimension
        :param batch_experts: list(batch) of lists(k) of up to k experts selected for this batch
        :returns: a tensor of scores, float32[batch_size, k]
        :note: if some rows in batch have less than max number of experts, their scores will be padded with -inf
        """
        expert_counts = list(map(len, batch_experts))
        batch_size = len(batch_experts)
        max_num_experts = max(expert_counts)
        total_num_experts = sum(expert_counts)

        device = grid_scores[0].device

        expert_index_in_batch = torch.arange(total_num_experts, device=device)
        expert_strides = torch.cumsum(torch.as_tensor([0] + expert_counts, device=device), dim=-1)[:-1]
        flat_batch_indices = (expert_index_in_batch >= expert_strides[:, None]).to(torch.int32).sum(0) - 1
        flat_local_indices = expert_index_in_batch - expert_strides[flat_batch_indices]
        flat_experts = [expert for row in batch_experts for expert in row]

        grid_indices = torch.zeros([len(flat_experts), len(grid_scores)], dtype=torch.int64)
        for i, expert in enumerate(flat_experts):
            expert_indices = expert.uid[len(self.beam_search.uid_prefix) :]
            expert_indices = list(map(int, expert_indices.split(UID_DELIMITER)))
            grid_indices[i] = torch.as_tensor(expert_indices, dtype=grid_indices.dtype)

        scores_per_dim = [
            dim_scores[flat_batch_indices, dim_indices] if len(flat_batch_indices) else torch.zeros(0, device=device)
            for dim_scores, dim_indices in zip(grid_scores, grid_indices.T)
        ]
        flat_scores = torch.sum(torch.stack(scores_per_dim, dim=0), dim=0)

        scores = torch.full((batch_size, max_num_experts), fill_value=-float("inf"), device=device)
        scores[flat_batch_indices, flat_local_indices] = flat_scores  # backprop-able w.r.t. flat_scores
        return scores

    @property
    def info(self):
        if self._expert_info is None:
            # grab some expert to set ensemble output shape
            proj_device = self.proj.weight.device
            dummy_scores_concat = self.proj(torch.randn(1, self.proj.in_features, device=proj_device))
            dummy_scores = dummy_scores_concat.cpu().detach().split_with_sizes(self.beam_search.grid_size, dim=-1)
            dummy_experts = self.beam_search.find_best_experts(dummy_scores, beam_size=1)
            self._expert_info = dummy_experts[0].info
        return self._expert_info


class _RemoteCallMany(torch.autograd.Function):
    """
    Internal autograd-friendly function that calls multiple experts on a batch of inputs and awaits responses
    This function that can recover from individual failures during forward and/or backward pass as long as at least
    one expert succeeds for each input. For user-friendly version of this function, use RemoteMixtureOfExperts module.

    Note: experts that failed during forward will be assigned zero outputs and marked as mask[i, j] = 0,
          experts that failed during backward will be treated as constants (i.e. gradients through them are zeros)
    """

    @classmethod
    def forward(
        cls,
        ctx,
        dummy,
        experts_per_sample: List[List[RemoteExpert]],
        k_min: int,
        backward_k_min: int,
        timeout_after_k_min: float,
        forward_timeout: Optional[float],
        backward_timeout: Optional[float],
        detect_anomalies: bool,
        allow_zero_outputs: bool,
        info: Dict[str, Any],
        *flat_inputs: torch.Tensor,
    ) -> Tuple[torch.Tensor]:
        assert not torch.is_grad_enabled()
        num_samples, max_experts = len(experts_per_sample), max(map(len, experts_per_sample))

        flat_inputs_cpu = []
        for tensor in flat_inputs:
            if detect_anomalies and not tensor.isfinite().all():
                raise ValueError("One of inputs has nan/inf values")
            flat_inputs_cpu.append(tensor.cpu())

        flat_inputs_per_sample = list(zip(*(x.split(1, dim=0) for x in flat_inputs_cpu)))
        assert len(experts_per_sample) == len(flat_inputs_per_sample) == num_samples

        # dispatch tasks to all remote experts collect responses
        pending_tasks: Dict[Future, Tuple[int, int]] = {}
        for i in range(num_samples):
            for j, expert in enumerate(experts_per_sample[i]):
                stub = get_server_stub(expert.p2p, expert.peer_id)
                serialized_tensors = (
                    serialize_torch_tensor(tensor, proto.compression)
                    for tensor, proto in zip(flat_inputs_per_sample[i], nested_flatten(info["forward_schema"]))
                )
                new_task = RemoteExpertWorker.run_coroutine(
                    expert_forward(expert.uid, flat_inputs_per_sample[i], serialized_tensors, stub),
                    return_future=True,
                )
                pending_tasks[new_task] = (i, j)

        responded_inds, alive_flat_outputs = cls._collect_responses(
            pending_tasks, num_samples, k_min, forward_timeout, timeout_after_k_min, detect_anomalies
        )
        if len(responded_inds) < k_min:
            raise TimeoutError(f"Forward pass: less than {k_min} responded within timeout")

        if not isinstance(info["outputs_schema"], tuple):
            outputs_schema = (info["outputs_schema"],)
        else:
            outputs_schema = info["outputs_schema"]
        outputs = nested_map(
            lambda descriptor: descriptor.make_zeros(num_samples, max_experts, device=flat_inputs[0].device),
            outputs_schema,
        )

        # assemble responses
        if len(responded_inds) > 0 or allow_zero_outputs:
            batch_inds, expert_inds = map(
                lambda x: torch.as_tensor(x, device=flat_inputs[0].device, dtype=torch.long),
                list(zip(*responded_inds)) or ([], []),
            )

            alive_flat_outputs_stacked = (torch.cat(outputs) for outputs in zip(*alive_flat_outputs))
            # torch tensors, i-th tensor is of shape [num_responded, *expert_outputs[i].shape]

            for output, response_stacked in zip(outputs, alive_flat_outputs_stacked):
                output[batch_inds, expert_inds] = response_stacked.to(output.device)

        else:
            raise RuntimeError("Forward pass: 0 experts responded within timeout and allow_zero_outputs is False")

        mask = torch.zeros([num_samples, max_experts], dtype=torch.bool, device=flat_inputs[0].device)
        mask[batch_inds, expert_inds] = True

        # save individual outputs for backward pass
        ctx.save_for_backward(batch_inds, expert_inds, *flat_inputs_cpu)
        ctx._saved_non_tensors = (
            info,
            backward_k_min,
            backward_timeout,
            timeout_after_k_min,
            experts_per_sample,
            detect_anomalies,
        )

        return (mask,) + outputs

    @classmethod
    @once_differentiable
    def backward(cls, ctx, *raw_grads: torch.Tensor) -> Tuple[Optional[torch.Tensor], ...]:
        assert not torch.is_grad_enabled()
        (
            info,
            backward_k_min,
            backward_timeout,
            timeout_after_k_min,
            expert_per_sample,
            detect_anomalies,
        ) = ctx._saved_non_tensors
        alive_ii, alive_jj, *flat_inputs_cpu = ctx.saved_tensors

        dummy_grad_mask, *flat_grad_outputs = raw_grads

        flat_grad_outputs_cpu = []
        for tensor in flat_grad_outputs:
            if detect_anomalies and not tensor.isfinite().all():
                raise ValueError("One of gradients has nan/inf values")
            flat_grad_outputs_cpu.append(tensor.cpu())

        num_samples, max_experts = dummy_grad_mask.shape

        inputs_per_expert = zip(*(tensor[alive_ii].split(1, dim=0) for tensor in flat_inputs_cpu))
        grad_outputs_per_expert = zip(
            *(tensor[alive_ii, alive_jj].split(1, dim=0) for tensor in flat_grad_outputs_cpu)
        )
        backward_schema = tuple(nested_flatten((info["forward_schema"], info["outputs_schema"])))

        # dispatch tasks to all remote experts, collect responses
        pending_tasks = {}
        for i, j, inputs_ij, grad_outputs_ij in zip(
            alive_ii.cpu().numpy(), alive_jj.cpu().numpy(), inputs_per_expert, grad_outputs_per_expert
        ):
            expert: RemoteExpert = expert_per_sample[i.item()][j.item()]
            stub = get_server_stub(expert.p2p, expert.peer_id)
            inputs_and_grad_outputs = tuple(nested_flatten((inputs_ij, grad_outputs_ij)))
            serialized_tensors = (
                serialize_torch_tensor(tensor, proto.compression)
                for tensor, proto in zip(inputs_and_grad_outputs, backward_schema)
            )
            new_task = RemoteExpertWorker.run_coroutine(
                expert_backward(expert.uid, inputs_and_grad_outputs, serialized_tensors, stub), return_future=True
            )
            pending_tasks[new_task] = (i, j)

        survivor_inds, survivor_grad_inputs = cls._collect_responses(
            pending_tasks, num_samples, backward_k_min, backward_timeout, timeout_after_k_min, detect_anomalies
        )
        if len(survivor_inds) < backward_k_min:
            raise TimeoutError(f"Backward pass: less than {backward_k_min} experts responded within timeout")

        # assemble responses
        batch_inds, expert_inds = map(
            lambda x: torch.as_tensor(x, dtype=torch.long), list(zip(*survivor_inds)) or ([], [])
        )

        survivor_grad_inputs_stacked = (torch.cat(grad_inputs) for grad_inputs in zip(*survivor_grad_inputs))
        # torch tensors, i-th tensor is of shape [num_backward_survivors, *flat_inputs_cpu[i].shape]

        grad_inputs = nested_map(
            lambda descr: descr.make_zeros(num_samples, device=flat_grad_outputs[0].device),
            list(nested_flatten(info["forward_schema"])),
        )

        for grad_input, survivor_grad_stacked in zip(grad_inputs, survivor_grad_inputs_stacked):
            grad_input_per_expert = torch.zeros(  # gradient tensor with individual contributions from each expert
                (num_samples, max_experts, *grad_input.shape[1:]),
                device=survivor_grad_stacked.device,
                dtype=survivor_grad_stacked.dtype,
            )
            grad_input_per_expert[batch_inds, expert_inds] = survivor_grad_stacked
            grad_input.copy_(grad_input_per_expert.to(flat_grad_outputs[0].device).sum(dim=1))

        return (DUMMY, None, None, None, None, None, None, None, None, None, *grad_inputs)

    @staticmethod
    def _collect_responses(
        task_to_indices: Dict[Future, Tuple[int, int]],
        num_samples: int,
        k_min: int,
        timeout_total: Optional[float],
        timeout_after_k_min: Optional[float],
        detect_anomalies: bool,
    ) -> Tuple[List[Tuple[int, int]], List[Tuple[torch.Tensor, ...]]]:
        """await up to k_min results and any result submitted within timeout_after_k_min, cancel stragglers"""
        timeout_total = float("inf") if timeout_total is None else timeout_total
        timeout_after_k_min = float("inf") if timeout_after_k_min is None else timeout_after_k_min
        num_successful_tasks = [0 for _ in range(num_samples)]
        pending_samples = num_samples  # samples for which we have less than k_min results
        finished_indices, finished_outputs = [], []
        t_finish = time.perf_counter() + timeout_total
        pending_tasks = set(task_to_indices.keys())
        finished_tasks = Queue()

        try:
            # the algorithm below is essentially futures.as_completed, but for grpc.Future
            for task in pending_tasks:
                task.add_done_callback(finished_tasks.put)

            for _ in range(len(task_to_indices)):
                timeout = max(0.0, t_finish - time.perf_counter()) if t_finish != float("inf") else None
                task = finished_tasks.get(timeout=timeout)
                pending_tasks.discard(task)

                task_output = _process_dispatched_task(task, detect_anomalies)
                if task_output is not None:
                    finished_indices.append(task_to_indices[task])
                    finished_outputs.append(task_output)

                    # count how many successes we have for each input sample
                    sample_index = task_to_indices[task][0]
                    num_successful_tasks[sample_index] += 1
                    if num_successful_tasks[sample_index] == k_min:
                        pending_samples -= 1
                        if (
                            pending_samples <= 0
                        ):  # all tasks finished, await stragglers for at most timeout_after_k_min
                            t_finish = min(t_finish, time.perf_counter() + timeout_after_k_min)

        except Empty:
            pass  # we reached t_finish, this is normal behavior
        finally:
            for task in pending_tasks:
                task.cancel()
        return finished_indices, finished_outputs


def _process_dispatched_task(task: Future, detect_anomalies: bool) -> Optional[Tuple[torch.Tensor]]:
    if task.exception() or task.cancelled():
        logger.warning(f"Task {task} failed: {type(task.exception())}")
        return None

    outputs = task.result()
    for tensor in outputs:
        if detect_anomalies and not tensor.isfinite().all():
            logger.error(f"Task {task} failed: output tensor contains nan/inf values")
            return None

    return outputs



from hivemind.moe.client import RemoteExpert, RemoteMixtureOfExperts, RemoteSwitchMixtureOfExperts
from hivemind.moe.server import (
    ModuleBackend,
    Server,
    background_server,
    declare_experts,
    get_experts,
    register_expert_class,
)



import asyncio
from enum import Enum
from typing import AsyncIterator, Optional, Sequence, Set, Tuple, Type

import torch

from hivemind.averaging.partition import AllreduceException, BannedException, TensorPartContainer, TensorPartReducer
from hivemind.compression import deserialize_torch_tensor, serialize_torch_tensor
from hivemind.p2p import P2P, P2PContext, PeerID, ServicerBase, StubBase
from hivemind.proto import averaging_pb2
from hivemind.utils import get_logger
from hivemind.utils.asyncio import (
    achain,
    aiter_with_timeout,
    amap_in_executor,
    anext,
    as_aiter,
    attach_event_on_finished,
)

# flavour types
GroupID = bytes
logger = get_logger(__name__)


class AveragingMode(Enum):
    NODE = 0
    CLIENT = 1
    AUX = 2


class AllReduceRunner(ServicerBase):
    """
    An internal class that runs butterfly AllReduce in a predefined group of averagers.

    This class inherits hivemind.p2p.ServicerBase, so it can be used as an RPCServicer for testing purposes without
    creating a full DecentralizedAverager.

    :note: this class returns **differences** between averaged and local tensors in order to improve numerical stability
    :param p2p: a hivemind.p2p.P2P instance used for communication with other peers
    :param servicer_type: a hivemind.p2p.ServicerBase subclass whose RPC signatures are used
      when requesting other peers. Typically, it is DecentralizedAverager, its derivative,
      or AllReduceRunner itself (for testing purposes).
    :param prefix: namespace for servicer's RPCs (typically, equal to prefix for group keys)
    :param group_id: unique identifier of this specific all-reduce run
    :param tensors: local tensors that should be averaged with groupmates
    :param weight: scalar weight of this peer's tensors in the average (doesn't need to sum up to 1)
    :param peer_id: your peer_id, must be included in ordered_peer_ids
    :param ordered_peer_ids: group peer_ids ordered s.t. i-th peer_id is responsible for averaging i-th part
    :param peer_fractions: for each peer, a target fraction of vector elements that this peer should average
      (the actual number of values by peer will be nearly proportional, but there are no exact guarantees)
    :param modes: AveragingMode for each peer in ordered_peer_ids (normal, client-only or auxiliary)
    :param sender_timeout: during all_reduce, any sender that fails to send tensor chunk within this many seconds from
      previous chunk will be marked as failed and excluded from averaging. default: equal to next_chunk_timeout
    :param reducer_timeout: during all_reduce, any reducer that fails to send results chunk within this many seconds
      from previous chunk will be marked as failed and excluded from averaging. default: 2 x sender_timeout
    :param kwargs: additional parameters (e.g. part_size_bytes) will be passed to TensorPartContainer
    :note: Full-mode peers send and receive tensor parts concurrently, assuming a full-duplex TCP stream. In turn,
      non-averaging peers receive results only after they finish sending, which helps them avoid
      throughput issues in case of asymmetric high-latency connections (e.g. ACK compression).
    """

    def __init__(
        self,
        *,
        p2p: P2P,
        servicer_type: Type[ServicerBase],
        prefix: Optional[str],
        group_id: GroupID,
        tensors: Sequence[torch.Tensor],
        weight: Optional[float] = None,
        ordered_peer_ids: Sequence[PeerID],
        peer_fractions: Tuple[float, ...],
        modes: Optional[Sequence[AveragingMode]] = None,
        sender_timeout: Optional[float] = None,
        reducer_timeout: Optional[float] = None,
        **kwargs,
    ):
        self._p2p = p2p
        self.peer_id = p2p.peer_id
        assert self.peer_id in ordered_peer_ids, "peer_id is not a part of the group"
        if reducer_timeout is not None and (sender_timeout is None or reducer_timeout <= sender_timeout):
            raise ValueError(
                "If reducer_timeout is enabled, sender_timeout must be shorter than reducer_timeout. "
                "Otherwise, there is a chance that reducers will be banned while they await senders."
            )

        if not issubclass(servicer_type, ServicerBase):
            raise TypeError("`servicer_type` is expected to be a ServicerBase subclass")
        self._servicer_type = servicer_type
        self._prefix = prefix

        modes = modes or tuple(AveragingMode.CLIENT if frac == 0 else AveragingMode.NODE for frac in peer_fractions)
        assert len(modes) == len(ordered_peer_ids), "lists have inconsistent length"
        assert any(mode != AveragingMode.CLIENT for mode in modes), "cannot run allreduce without reducers"
        for mode, frac in zip(modes, peer_fractions):
            assert mode != AveragingMode.CLIENT or frac == 0, "client-mode peer should have zero all-reduce fraction"

        self.group_id, self.ordered_peer_ids = group_id, ordered_peer_ids
        self.modes, self.peer_fractions = modes, peer_fractions

        if weight is None:
            weight = float(modes[self.ordered_peer_ids.index(self.peer_id)] != AveragingMode.AUX)
        self.weight = weight

        self._future = asyncio.Future()

        self.sender_peer_ids = []
        for peer_id, mode in zip(self.ordered_peer_ids, modes):
            if mode != AveragingMode.AUX:
                self.sender_peer_ids.append(peer_id)

        self.sender_timeout, self.reducer_timeout = sender_timeout, reducer_timeout
        self.all_senders_started = asyncio.Event()
        self.banned_senders: Set[PeerID] = set()  # peers that did not send data by next_chunk_timeout
        self.banlock = asyncio.Lock()

        self.active_senders: Set[PeerID] = set()  # peers that began sending data via rpc_aggregate_part
        if self.peer_id in self.sender_peer_ids:
            self.active_senders.add(self.peer_id)
        if len(self.active_senders) == len(self.sender_peer_ids):
            self.all_senders_started.set()

        peer_id_index = self.ordered_peer_ids.index(self.peer_id)
        self.tensor_part_container = TensorPartContainer(tensors, peer_fractions, return_deltas=True, **kwargs)
        self.parts_for_local_averaging = self.tensor_part_container.get_raw_input_parts(peer_id_index)
        self.tensor_part_reducer = TensorPartReducer(
            tuple(part.shape for part in self.parts_for_local_averaging),
            len(self.sender_peer_ids),
        )

    def __repr__(self):
        return f"{self.__class__.__name__}({self.peer_id}, group_size={self.group_size})"

    def __aiter__(self):
        return self.run()

    def __contains__(self, peer_id: PeerID):
        return peer_id in self.ordered_peer_ids

    @property
    def group_size(self):
        return len(self.ordered_peer_ids)

    def _get_peer_stub(self, peer: PeerID) -> StubBase:
        return self._servicer_type.get_stub(self._p2p, peer, namespace=self._prefix)

    def should_delay_results(self, peer_id: PeerID) -> bool:
        return self.peer_fractions[self.ordered_peer_ids.index(peer_id)] == 0

    async def run(self) -> AsyncIterator[torch.Tensor]:
        """Run all-reduce, return differences between averaged and original tensors as they are computed"""
        pending_tasks = set()

        if self.tensor_part_container.num_parts_by_peer[self.ordered_peer_ids.index(self.peer_id)] != 0:
            pending_tasks.add(asyncio.create_task(self._handle_missing_senders()))

        try:
            if len(self.sender_peer_ids) == 0:
                logger.debug(f"{self} - finished all-reduce early: all peers are auxiliaries ({self.modes})")
                self.finalize()

            elif self.peer_id in self.sender_peer_ids:
                for peer_id, parts in zip(self.ordered_peer_ids, self.tensor_part_container.num_parts_by_peer):
                    if parts != 0:
                        pending_tasks.add(asyncio.create_task(self._communicate_with_peer(peer_id)))

                async for averaged_tensor_delta in self.tensor_part_container.iterate_output_tensors():
                    yield averaged_tensor_delta  # delta = averaged_tensor - original_tensor

                self.finalize()

            else:  # auxiliary peer
                await self.tensor_part_reducer.finished.wait()
                self.finalize()

        except BaseException as e:
            self.finalize(exception=e)
            for task in pending_tasks:
                task.cancel()
            raise

        finally:
            for task in pending_tasks:
                try:
                    await task
                except asyncio.CancelledError:
                    pass
                except Exception as inner_exc:
                    logger.debug(f"Task {task} failed with {inner_exc}", exc_info=True)

    async def _handle_missing_senders(self):
        """Detect senders that should have sent tensors for averaging, but did not send anything within timeout"""
        try:
            await asyncio.wait_for(self.all_senders_started.wait(), self.sender_timeout)
        except asyncio.TimeoutError:
            for peer_id in self.sender_peer_ids:
                if peer_id not in self.active_senders and peer_id not in self.banned_senders:
                    await self._ban_sender(peer_id)

    async def _communicate_with_peer(self, peer_id: PeerID):
        """Send a part of local tensors and metadata to a single peer, receive the average for that part of tensors"""
        peer_index = self.ordered_peer_ids.index(peer_id)
        if peer_id == self.peer_id:
            sender_index = self.sender_peer_ids.index(peer_id)
            for part_index, tensor_part in enumerate(self.parts_for_local_averaging):
                averaged_part = await self.tensor_part_reducer.accumulate_part(
                    sender_index, part_index, tensor_part, weight=self.weight
                )
                self.tensor_part_container.register_processed_part(peer_index, part_index, averaged_part - tensor_part)

        else:
            try:
                done_sending = asyncio.Event()
                inputs_aiter = attach_event_on_finished(self._generate_input_for_peer(peer_index), done_sending)
                stream = await self._get_peer_stub(peer_id).rpc_aggregate_part(inputs_aiter)

                if self.should_delay_results(self.peer_id):
                    await done_sending.wait()

                part_index = 0

                def _try_deserialize(msg):
                    if msg.code != averaging_pb2.AVERAGED_PART:
                        raise AllreduceException(f"{peer_id} sent {averaging_pb2.MessageCode.Name(msg.code)}")
                    return deserialize_torch_tensor(msg.tensor_part), msg

                async for delta, msg in amap_in_executor(
                    _try_deserialize,
                    aiter_with_timeout(stream, self.reducer_timeout),
                    max_prefetch=self.tensor_part_container.prefetch,
                ):
                    self.tensor_part_container.register_processed_part(peer_index, part_index, delta)
                    part_index += 1

                if part_index != self.tensor_part_container.num_parts_by_peer[peer_index]:
                    raise AllreduceException(
                        f"peer {peer_id} sent {part_index} parts, but we expected "
                        f"{self.tensor_part_container.num_parts_by_peer[peer_index]}"
                    )
            except BaseException as e:
                if isinstance(e, Exception):
                    logger.debug(f"Caught {repr(e)} when communicating to {peer_id}", exc_info=True)
                self.tensor_part_container.register_failed_reducer(peer_index)
                raise

    async def _generate_input_for_peer(self, peer_index: int) -> AsyncIterator[averaging_pb2.AveragingData]:
        parts_aiter = self.tensor_part_container.iterate_input_parts_for(peer_index)
        first_part = await anext(parts_aiter)
        yield averaging_pb2.AveragingData(
            code=averaging_pb2.PART_FOR_AVERAGING,
            group_id=self.group_id,
            tensor_part=first_part,
            weight=self.weight,
        )
        async for part in parts_aiter:
            yield averaging_pb2.AveragingData(tensor_part=part, weight=self.weight)

    async def rpc_aggregate_part(
        self, stream: AsyncIterator[averaging_pb2.AveragingData], context: P2PContext
    ) -> AsyncIterator[averaging_pb2.AveragingData]:
        """a peer sends us a part of his tensor; we should average it with other peers and return the difference"""
        sender_index = self.sender_peer_ids.index(context.remote_id)
        self.active_senders.add(context.remote_id)
        if len(self.active_senders) == len(self.sender_peer_ids):
            self.all_senders_started.set()

        try:
            request: averaging_pb2.AveragingData = await asyncio.wait_for(anext(stream), self.sender_timeout)
            reason_to_reject = self._check_reasons_to_reject(request, context)
            if reason_to_reject:
                yield reason_to_reject
                return

            elif request.code == averaging_pb2.PART_FOR_AVERAGING:
                stream = aiter_with_timeout(achain(as_aiter(request), stream), self.sender_timeout)
                if not self.should_delay_results(context.remote_id):
                    async for msg in self._accumulate_parts_streaming(stream, sender_index):
                        yield msg

                else:
                    done_receiving = asyncio.Event()
                    delayed_results = asyncio.Queue()

                    async def _accumulate_parts():
                        try:
                            async for msg in self._accumulate_parts_streaming(
                                attach_event_on_finished(stream, done_receiving), sender_index
                            ):
                                delayed_results.put_nowait(msg)
                        finally:
                            delayed_results.put_nowait(None)

                    accumulate_task = asyncio.create_task(_accumulate_parts())

                    await done_receiving.wait()

                    while True:
                        next_result = await delayed_results.get()
                        if next_result is None:
                            break
                        yield next_result
                    await accumulate_task

            else:
                yield averaging_pb2.AveragingData(code=averaging_pb2.INTERNAL_ERROR)
                raise AllreduceException(f"{context.remote_id} sent {averaging_pb2.MessageCode.Name(request.code)}")

        except BaseException as e:
            await self._ban_sender(context.remote_id)
            if isinstance(e, Exception):
                logger.debug(f"Caught {repr(e)} when communicating with {context.remote_id}", exc_info=True)
                yield averaging_pb2.AveragingData(code=averaging_pb2.INTERNAL_ERROR)
            else:
                raise  # CancelledError, StopIteration and similar

    async def _ban_sender(self, peer_id: PeerID):
        async with self.banlock:
            if peer_id not in self.banned_senders:
                self.banned_senders.add(peer_id)
                self.tensor_part_reducer.on_sender_failed(self.sender_peer_ids.index(peer_id))

    def _check_reasons_to_reject(
        self, request: averaging_pb2.AveragingData, context: P2PContext
    ) -> Optional[averaging_pb2.AveragingData]:
        if request.group_id != self.group_id:
            return averaging_pb2.AveragingData(code=averaging_pb2.BAD_GROUP_ID)
        elif self._future.cancelled():
            return averaging_pb2.AveragingData(code=averaging_pb2.CANCELLED)
        elif self._future.done():
            return averaging_pb2.AveragingData(code=averaging_pb2.INTERNAL_ERROR)
        elif context.remote_id not in self.sender_peer_ids:
            return averaging_pb2.AveragingData(code=averaging_pb2.PROTOCOL_VIOLATION)

    async def _accumulate_parts_streaming(self, stream: AsyncIterator[averaging_pb2.AveragingData], sender_index: int):
        part_index = 0
        try:
            loop = asyncio.get_event_loop()
            async for tensor_part, weight, part_compression in amap_in_executor(
                lambda msg: (deserialize_torch_tensor(msg.tensor_part), msg.weight, msg.tensor_part.compression),
                stream,
                max_prefetch=self.tensor_part_container.prefetch,
            ):
                try:
                    averaged_part = await self.tensor_part_reducer.accumulate_part(
                        sender_index, part_index, tensor_part, weight=weight
                    )
                    part_index += 1
                except BannedException:
                    logger.debug(f"Sender {sender_index} is already banned")
                    break  # sender was banned, we no longer need to aggregate it

                serialized_delta = await loop.run_in_executor(
                    None, lambda: serialize_torch_tensor(averaged_part - tensor_part, part_compression)
                )
                yield averaging_pb2.AveragingData(code=averaging_pb2.AVERAGED_PART, tensor_part=serialized_delta)
        finally:
            if part_index != self.tensor_part_reducer.num_parts:
                await self._ban_sender(self.sender_peer_ids[sender_index])

    def finalize(self, *, cancel: bool = False, exception: Optional[BaseException] = None):
        """finish or terminate AllReduceRunner, propagate any errors / cancellations to peers."""
        assert not cancel or not exception, "finalize accepts either exception or cancel, but not both"
        if not self._future.done():
            if cancel:
                logger.debug(f"{self} - cancelled")
                self._future.cancel()
            elif exception:
                logger.debug(f"{self} - caught {exception}")
                self._future.set_exception(exception)
            else:
                logger.debug(f"{self} - finished")
                self._future.set_result(None)
            self.tensor_part_container.finalize()
            self.tensor_part_reducer.finalize()
        else:
            logger.debug(f"{self} - attempted to finalize allreduce that is already finished: {self._future}")



""" A dictionary-like storage that stores items until a specified expiration time or up to a limited size """
from __future__ import annotations

import heapq
import time
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Dict, Generic, Iterator, List, Optional, Tuple, TypeVar

KeyType = TypeVar("KeyType")
ValueType = TypeVar("ValueType")
get_dht_time = time.time  # a global (weakly synchronized) time
MAX_DHT_TIME_DISCREPANCY_SECONDS = 3  # max allowed difference between get_dht_time for two DHT nodes
DHTExpiration = float
ROOT = 0


@dataclass(init=True, repr=True, frozen=True)
class ValueWithExpiration(Generic[ValueType]):
    value: ValueType
    expiration_time: DHTExpiration

    def __iter__(self):
        return iter((self.value, self.expiration_time))

    def __getitem__(self, item):
        if item == 0:
            return self.value
        elif item == 1:
            return self.expiration_time
        else:
            return getattr(self, item)

    def __eq__(self, item):
        if isinstance(item, ValueWithExpiration):
            return self.value == item.value and self.expiration_time == item.expiration_time
        elif isinstance(item, tuple):
            return tuple.__eq__((self.value, self.expiration_time), item)
        else:
            return False


@dataclass(init=True, repr=True, order=True, frozen=True)
class HeapEntry(Generic[KeyType]):
    expiration_time: DHTExpiration
    key: KeyType


class TimedStorage(Generic[KeyType, ValueType]):
    """A dictionary that maintains up to :maxsize: key-value-expiration tuples until their expiration_time"""

    frozen = False  # can be set to True. If true, do not remove outdated elements

    def __init__(self, maxsize: Optional[int] = None):
        self.maxsize = maxsize or float("inf")
        self.data: Dict[KeyType, ValueWithExpiration[ValueType]] = dict()
        self.expiration_heap: List[HeapEntry[KeyType]] = []
        self.key_to_heap: Dict[KeyType, HeapEntry[KeyType]] = dict()

    def _remove_outdated(self):
        while (
            not self.frozen
            and self.expiration_heap
            and (self.expiration_heap[ROOT].expiration_time < get_dht_time() or len(self.data) > self.maxsize)
        ):
            heap_entry = heapq.heappop(self.expiration_heap)
            if self.key_to_heap.get(heap_entry.key) == heap_entry:
                del self.data[heap_entry.key], self.key_to_heap[heap_entry.key]

    def store(self, key: KeyType, value: ValueType, expiration_time: DHTExpiration) -> bool:
        """
        Store a (key, value) pair locally at least until expiration_time. See class docstring for details.
        :returns: True if new value was stored, False it was rejected (current value is newer)
        """
        if expiration_time < get_dht_time() and not self.frozen:
            return False
        self.key_to_heap[key] = HeapEntry(expiration_time, key)
        heapq.heappush(self.expiration_heap, self.key_to_heap[key])
        if key in self.data:
            if self.data[key].expiration_time < expiration_time:
                self.data[key] = ValueWithExpiration(value, expiration_time)
                return True
            return False
        self.data[key] = ValueWithExpiration(value, expiration_time)
        self._remove_outdated()
        return True

    def get(self, key: KeyType) -> Optional[ValueWithExpiration[ValueType]]:
        """Get a value corresponding to a key if that (key, value) pair was previously stored under this key."""
        self._remove_outdated()
        if key in self.data:
            return self.data[key]
        return None

    def items(self) -> Iterator[Tuple[KeyType, ValueWithExpiration[ValueType]]]:
        """Iterate over (key, value, expiration_time) tuples stored in this storage"""
        self._remove_outdated()
        return ((key, value_and_expiration) for key, value_and_expiration in self.data.items())

    def top(self) -> Tuple[Optional[KeyType], Optional[ValueWithExpiration[ValueType]]]:
        """Return the entry with earliest expiration or None if there isn't any"""
        self._remove_outdated()
        if self.data:
            # skip leftover "ghost" entries until first real entry
            while self.key_to_heap.get(self.expiration_heap[ROOT].key) != self.expiration_heap[ROOT]:
                heapq.heappop(self.expiration_heap)
            top_key = self.expiration_heap[ROOT].key
            return top_key, self.data[top_key]
        return None, None

    def clear(self):
        self.data.clear()
        self.key_to_heap.clear()
        self.expiration_heap.clear()

    def __contains__(self, key: KeyType):
        self._remove_outdated()
        return key in self.data

    def __len__(self):
        self._remove_outdated()
        return len(self.data)

    def __delitem__(self, key: KeyType):
        if key in self.key_to_heap:
            del self.data[key], self.key_to_heap[key]
        # note: key may still be in self.expiration_heap, but it will not be used and eventually ._remove_outdated()

    def __bool__(self):
        return bool(self.data)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.data})"

    @contextmanager
    def freeze(self):
        """Temporarily cease to ._remove_outdated() elements inside this context to ensure consistency"""
        prev_frozen, self.frozen = self.frozen, True
        try:
            yield self
        finally:
            self.frozen = prev_frozen



from __future__ import annotations

from concurrent.futures import Future
from typing import Any, Dict, Iterable, List, Optional, Sequence, Tuple, Union

import torch
import torch.nn as nn
from torch.autograd.function import once_differentiable

from hivemind import moe
from hivemind.compression import deserialize_tensor_stream, deserialize_torch_tensor, serialize_torch_tensor
from hivemind.dht import DHT
from hivemind.moe.client.remote_expert_worker import RemoteExpertWorker
from hivemind.moe.expert_uid import ExpertInfo
from hivemind.p2p import P2P, PeerID, StubBase
from hivemind.p2p.p2p_daemon_bindings.control import DEFAULT_MAX_MSG_SIZE, MAX_UNARY_PAYLOAD_SIZE
from hivemind.proto import runtime_pb2
from hivemind.utils.asyncio import amap_in_executor, iter_as_aiter
from hivemind.utils.mpfuture import MPFuture
from hivemind.utils.nested import nested_compare, nested_flatten, nested_pack
from hivemind.utils.serializer import MSGPackSerializer
from hivemind.utils.streaming import split_for_streaming

DUMMY = torch.empty(0, requires_grad=True)  # dummy tensor that triggers autograd in RemoteExpert


def get_server_stub(p2p: P2P, server_peer_id: PeerID) -> "ConnectionHandlerStub":
    """Create an RPC stub that can send requests to any expert on the specified remote server"""
    return moe.server.connection_handler.ConnectionHandler.get_stub(p2p, server_peer_id)


class RemoteExpert(nn.Module):
    """
    A simple module that runs forward/backward of an expert hosted on a remote machine.
    Works seamlessly with pytorch autograd. (this is essentially a simple RPC function)
    Warning: RemoteExpert currently assumes that you provide it with correct input shapes.
    Sending wrong input shapes can cause RemoteExpert to freeze indefinitely due to error in runtime.

    :param expert_info: RemoteExpertInfo with uid and server PeerInfo
    :param p2p: P2P instance connected to the running p2pd
    """

    def __init__(self, expert_info: ExpertInfo, p2p: P2P):
        super().__init__()
        self._info, self.p2p = expert_info, p2p
        self._rpc_info = None

    @property
    def uid(self):
        return self._info.uid

    @property
    def peer_id(self) -> PeerID:
        return self._info.peer_id

    @property
    def stub(self) -> StubBase:
        return get_server_stub(self.p2p, self.peer_id)

    def forward(self, *args, **kwargs):
        """Call RemoteExpert for the specified inputs and return its output(s). Compatible with pytorch.autograd."""
        assert len(kwargs) == len(self.info["keyword_names"]), f"Keyword args should be {self.info['keyword_names']}"
        kwargs = {key: kwargs[key] for key in self.info["keyword_names"]}

        # Note: we put keyword arguments in the same order as on a server to prevent f(a=1, b=2) != f(b=2, a=1) errors

        forward_inputs = (args, kwargs)

        if not nested_compare(forward_inputs, self.info["forward_schema"]):
            raise TypeError(f"Inputs do not match expert input schema. Did you pass the right number of parameters?")

        flat_outputs = _RemoteModuleCall.apply(DUMMY, self.uid, self.stub, self.info, *nested_flatten(forward_inputs))

        # Note: we send DUMMY to prevent torch from excluding expert from backward if no other inputs require grad
        return nested_pack(flat_outputs, structure=self.info["outputs_schema"])

    @property
    def info(self):
        if self._rpc_info is None:
            outputs = RemoteExpertWorker.run_coroutine(self.stub.rpc_info(runtime_pb2.ExpertUID(uid=self.uid)))
            self._rpc_info = MSGPackSerializer.loads(outputs.serialized_info)
        return self._rpc_info

    def extra_repr(self):
        return f"uid={self.uid}, server_peer_id={self.peer_id}"


def _create_remote_experts(infos: Sequence[Optional[ExpertInfo]], p2p: P2P) -> List[Optional[RemoteExpert]]:
    experts: List[Optional[RemoteExpert]] = []
    for info in infos:
        if info is not None:
            experts.append(RemoteExpert(info, p2p))
        else:
            experts.append(None)
    return experts


def create_remote_experts(
    infos: Union[Sequence[Optional[ExpertInfo]], MPFuture], dht: DHT, return_future: bool = False
) -> Union[List[Optional[RemoteExpert]], Future]:
    if return_future:

        async def _unpack(infos_future: MPFuture, dht: DHT):
            p2p = await dht.replicate_p2p()
            return _create_remote_experts(await infos_future, p2p)

        return RemoteExpertWorker.run_coroutine(_unpack(infos, dht), return_future)

    p2p = RemoteExpertWorker.run_coroutine(dht.replicate_p2p())
    return _create_remote_experts(infos, p2p)


def batch_create_remote_experts(
    infos: Union[Sequence[Sequence[Optional[ExpertInfo]]], MPFuture],
    dht: DHT,
    return_future: bool = False,
) -> Union[List[List[Optional[RemoteExpert]]], Future]:
    if return_future:

        async def _unpack(infos_future: MPFuture, dht: DHT):
            p2p = await dht.replicate_p2p()
            return [_create_remote_experts(i, p2p) for i in await infos_future]

        return RemoteExpertWorker.run_coroutine(_unpack(infos, dht), return_future)

    return [create_remote_experts(exps, dht) for exps in infos]


async def _backward_stream(uid: str, serialized_tensors: Iterable[runtime_pb2.Tensor], stub) -> List[torch.Tensor]:
    split = (part for tensor in serialized_tensors for part in split_for_streaming(tensor, DEFAULT_MAX_MSG_SIZE))

    grad_inputs = await stub.rpc_backward_stream(
        amap_in_executor(
            lambda tensor: runtime_pb2.ExpertRequest(uid=uid, tensors=[tensor]),
            iter_as_aiter(split),
        ),
    )
    tensors_stream = amap_in_executor(lambda msg: msg.tensors, grad_inputs)
    return await deserialize_tensor_stream(tensors_stream)


async def _backward_unary(uid: str, serialized_tensors: Iterable[runtime_pb2.Tensor], stub) -> List[torch.Tensor]:
    grad_inputs: runtime_pb2.ExpertResponse = await stub.rpc_backward(
        runtime_pb2.ExpertRequest(uid=uid, tensors=list(serialized_tensors))
    )
    return [deserialize_torch_tensor(t) for t in grad_inputs.tensors]


async def expert_backward(
    uid: str, inputs_and_grads: Sequence[torch.Tensor], serialized_tensors: Iterable[runtime_pb2.Tensor], stub
) -> List[torch.Tensor]:
    size = 0
    for t in inputs_and_grads:
        size += t.element_size() * t.nelement()
        if size > MAX_UNARY_PAYLOAD_SIZE:
            return await _backward_stream(uid, serialized_tensors, stub)
    else:
        return await _backward_unary(uid, serialized_tensors, stub)


async def _forward_stream(uid: str, serialized_tensors: Iterable[runtime_pb2.Tensor], stub) -> List[torch.Tensor]:
    split = (p for t in serialized_tensors for p in split_for_streaming(t, DEFAULT_MAX_MSG_SIZE))

    outputs = await stub.rpc_forward_stream(
        amap_in_executor(
            lambda tensor: runtime_pb2.ExpertRequest(uid=uid, tensors=[tensor]),
            iter_as_aiter(split),
        ),
    )

    tensors_stream = amap_in_executor(lambda msg: msg.tensors, outputs)
    return await deserialize_tensor_stream(tensors_stream)


async def _forward_unary(uid: str, serialized_tensors: Iterable[runtime_pb2.Tensor], stub) -> List[torch.Tensor]:
    outputs: runtime_pb2.ExpertResponse = await stub.rpc_forward(
        runtime_pb2.ExpertRequest(uid=uid, tensors=list(serialized_tensors))
    )
    return [deserialize_torch_tensor(t) for t in outputs.tensors]


async def expert_forward(
    uid: str, inputs: Sequence[torch.Tensor], serialized_tensors: Iterable[runtime_pb2.Tensor], stub
) -> List[torch.Tensor]:
    size = 0
    for t in inputs:
        size += t.element_size() * t.nelement()
        if size > MAX_UNARY_PAYLOAD_SIZE:
            return await _forward_stream(uid, serialized_tensors, stub)
    else:
        return await _forward_unary(uid, serialized_tensors, stub)


class _RemoteModuleCall(torch.autograd.Function):
    """Internal autograd-friendly call of a remote module. For applications, use RemoteExpert instead."""

    @staticmethod
    def forward(
        ctx,
        dummy: torch.Tensor,
        uid: str,
        stub: "ConnectionHandlerStub",
        info: Dict[str, Any],
        *inputs: torch.Tensor,
    ) -> Tuple[torch.Tensor, ...]:
        # Note: *inputs are flattened input tensors that follow the expert's info['input_schema']
        # detach to avoid pickling the computation graph
        inputs = tuple(tensor.cpu().detach() for tensor in inputs)
        ctx.uid, ctx.stub, ctx.info = uid, stub, info
        ctx.save_for_backward(*inputs)
        serialized_tensors = (
            serialize_torch_tensor(tensor, proto.compression)
            for tensor, proto in zip(inputs, nested_flatten(info["forward_schema"]))
        )
        deserialized_outputs = RemoteExpertWorker.run_coroutine(expert_forward(uid, inputs, serialized_tensors, stub))

        return tuple(deserialized_outputs)

    @staticmethod
    @once_differentiable
    def backward(ctx, *grad_outputs) -> Tuple[Optional[torch.Tensor], ...]:
        grad_outputs_cpu = tuple(tensor.cpu() for tensor in grad_outputs)
        inputs_and_grad_outputs = tuple(nested_flatten((ctx.saved_tensors, grad_outputs_cpu)))
        backward_schema = tuple(nested_flatten((ctx.info["forward_schema"], ctx.info["outputs_schema"])))
        serialized_tensors = (
            serialize_torch_tensor(tensor, proto.compression)
            for tensor, proto in zip(inputs_and_grad_outputs, backward_schema)
        )
        deserialized_grad_inputs = RemoteExpertWorker.run_coroutine(
            expert_backward(ctx.uid, inputs_and_grad_outputs, serialized_tensors, ctx.stub)
        )

        return (DUMMY, None, None, None, *deserialized_grad_inputs)



import random
import re
from typing import List, Optional, Tuple

import numpy as np

from hivemind.averaging.group_info import GroupInfo
from hivemind.dht import DHT
from hivemind.p2p import PeerID
from hivemind.utils import DHTExpiration, get_logger

GroupKey = str
GROUP_PATTERN = re.compile("^(([^.])+)[.]0b[01]*$")  # e.g. bert_exp4_averaging.0b01001101
DEFAULT_NUM_BUCKETS = 256
logger = get_logger(__name__)


def is_valid_group(maybe_group: str) -> bool:
    """A group identifier must contain group type, followed by one or more .-separated indices, and any ?metadata"""
    return bool(GROUP_PATTERN.fullmatch(maybe_group))


class GroupKeyManager:
    """
    Utility class that declares and fetches averaging-related keys using a DHT
    """

    def __init__(
        self,
        dht: DHT,
        prefix: str,
        initial_group_bits: str,
        target_group_size: Optional[int],
    ):
        assert all(bit in "01" for bit in initial_group_bits)
        if target_group_size is not None and not is_power_of_two(target_group_size):
            logger.warning("It is recommended to set target_group_size to a power of 2")

        self.dht, self.prefix, self.group_bits = dht, prefix, initial_group_bits
        self.target_group_size = target_group_size
        self.peer_id = dht.peer_id

    @property
    def current_key(self) -> GroupKey:
        return f"{self.prefix}.0b{self.group_bits}"

    async def declare_averager(
        self, group_key: GroupKey, peer_id: PeerID, expiration_time: float, looking_for_group: bool = True
    ) -> bool:
        """
        Add (or remove) the averager to a given allreduce bucket

        :param group_key: allreduce group key, e.g. my_averager.0b011011101
        :param peer_id: averager public peer_id for incoming requests
        :param expiration_time: intent to run allreduce before this timestamp
        :param looking_for_group: by default (True), declare the averager as "looking for group" in a given group;
          If False, this will instead mark that the averager as no longer looking for group, (e.g. it already finished)
        :return: True if declared, False if declaration was rejected by DHT peers
        :note: when leaving (i.e. is_active=False), please specify the same expiration_time as when entering the group
        :note: setting is_active=False does *not* guarantee that others will immediately stop to query you.
        """
        expiration_time = expiration_time if looking_for_group else float(np.nextafter(expiration_time, float("inf")))
        return await self.dht.store(
            key=group_key,
            subkey=peer_id.to_bytes(),
            value=looking_for_group,
            expiration_time=expiration_time,
            return_future=True,
        )

    async def get_averagers(self, group_key: GroupKey, only_active: bool) -> List[Tuple[PeerID, DHTExpiration]]:
        """
        Find and return averagers that were declared with a given all-reduce key

        :param group_key: finds averagers that have the this group key, e.g. my_averager.0b011011101
        :param only_active: if True, return only active averagers that are looking for group (i.e. with value = True)
            if False, return all averagers under a given group_key regardless of value
        :return: peer_ids and expirations of every matching averager
        """
        assert is_valid_group(group_key), f"Group key {group_key} is invalid, must follow {GROUP_PATTERN}"
        result = await self.dht.get(group_key, latest=True, return_future=True)
        if result is None or not isinstance(result.value, dict):
            logger.debug(f"Allreduce group not found: {group_key}, creating new group")
            return []
        averagers = []
        for key, looking_for_group in result.value.items():
            try:
                if only_active and not looking_for_group.value:
                    continue
                averagers.append((PeerID(key), looking_for_group.expiration_time))
            except Exception as e:
                logger.warning(f"Could not parse group key {key} ({looking_for_group}, exc={e})")
        return averagers

    async def update_key_on_group_assembled(self, group_info: GroupInfo, is_leader: bool = True):
        """this function is triggered every time an averager finds an allreduce group"""
        rng = random.Random(group_info.group_id)
        index = group_info.peer_ids.index(self.peer_id)
        num_buckets = self.target_group_size
        if num_buckets is None:
            num_buckets = next_power_of_two(group_info.group_size)
        generalized_index = rng.sample(range(num_buckets), group_info.group_size)[index]
        nbits = int(np.ceil(np.log2(num_buckets)))
        new_bits = bin(generalized_index)[2:].rjust(nbits, "0")
        self.group_bits = (self.group_bits + new_bits)[-len(self.group_bits) :] if self.group_bits else ""
        logger.debug(f"{self.peer_id} - updated group key to {self.group_bits}")

    async def update_key_on_not_enough_peers(self):
        """this function is triggered whenever averager fails to assemble group within timeout"""
        pass  # to be implemented in subclasses


def is_power_of_two(n):
    """Check whether n is a power of 2"""
    return (n != 0) and (n & (n - 1) == 0)


def next_power_of_two(n):
    """Round n up to the nearest power of 2"""
    return 1 if n == 0 else 2 ** (n - 1).bit_length()



import asyncio
import functools
import secrets
from abc import ABC, abstractmethod
from datetime import timedelta
from enum import Enum
from typing import Optional

from hivemind.proto.auth_pb2 import AccessToken, RequestAuthInfo, ResponseAuthInfo
from hivemind.utils.crypto import RSAPrivateKey, RSAPublicKey
from hivemind.utils.logging import get_logger
from hivemind.utils.timed_storage import TimedStorage, get_dht_time

logger = get_logger(__name__)


class AuthorizedRequestBase:
    """
    Interface for protobufs with the ``RequestAuthInfo auth`` field. Used for type annotations only.
    """

    auth: RequestAuthInfo


class AuthorizedResponseBase:
    """
    Interface for protobufs with the ``ResponseAuthInfo auth`` field. Used for type annotations only.
    """

    auth: ResponseAuthInfo


class AuthorizerBase(ABC):
    @abstractmethod
    async def sign_request(self, request: AuthorizedRequestBase, service_public_key: Optional[RSAPublicKey]) -> None:
        ...

    @abstractmethod
    async def validate_request(self, request: AuthorizedRequestBase) -> bool:
        ...

    @abstractmethod
    async def sign_response(self, response: AuthorizedResponseBase, request: AuthorizedRequestBase) -> None:
        ...

    @abstractmethod
    async def validate_response(self, response: AuthorizedResponseBase, request: AuthorizedRequestBase) -> bool:
        ...


class TokenAuthorizerBase(AuthorizerBase):
    """
    Implements the authorization protocol for a moderated Hivemind network.
    See https://github.com/learning-at-home/hivemind/issues/253
    """

    def __init__(self, local_private_key: Optional[RSAPrivateKey] = None):
        if local_private_key is None:
            local_private_key = RSAPrivateKey.process_wide()
        self._local_private_key = local_private_key
        self._local_public_key = local_private_key.get_public_key()

        self._local_access_token = None
        self._refresh_lock = asyncio.Lock()

        self._recent_nonces = TimedStorage()

    @abstractmethod
    async def get_token(self) -> AccessToken:
        ...

    @abstractmethod
    def is_token_valid(self, access_token: AccessToken) -> bool:
        ...

    @abstractmethod
    def does_token_need_refreshing(self, access_token: AccessToken) -> bool:
        ...

    async def refresh_token_if_needed(self) -> None:
        if self._local_access_token is None or self.does_token_need_refreshing(self._local_access_token):
            async with self._refresh_lock:
                if self._local_access_token is None or self.does_token_need_refreshing(self._local_access_token):
                    self._local_access_token = await self.get_token()
                    assert self.is_token_valid(self._local_access_token)

    @property
    def local_public_key(self) -> RSAPublicKey:
        return self._local_public_key

    async def sign_request(self, request: AuthorizedRequestBase, service_public_key: Optional[RSAPublicKey]) -> None:
        await self.refresh_token_if_needed()
        auth = request.auth

        auth.client_access_token.CopyFrom(self._local_access_token)

        if service_public_key is not None:
            auth.service_public_key = service_public_key.to_bytes()
        auth.time = get_dht_time()
        auth.nonce = secrets.token_bytes(8)

        assert auth.signature == b""
        auth.signature = self._local_private_key.sign(request.SerializeToString())

    _MAX_CLIENT_SERVICER_TIME_DIFF = timedelta(minutes=1)

    async def validate_request(self, request: AuthorizedRequestBase) -> bool:
        await self.refresh_token_if_needed()
        auth = request.auth

        if not self.is_token_valid(auth.client_access_token):
            logger.debug("Client failed to prove that it (still) has access to the network")
            return False

        client_public_key = RSAPublicKey.from_bytes(auth.client_access_token.public_key)
        signature = auth.signature
        auth.signature = b""
        if not client_public_key.verify(request.SerializeToString(), signature):
            logger.debug("Request has invalid signature")
            return False

        if auth.service_public_key and auth.service_public_key != self._local_public_key.to_bytes():
            logger.debug("Request is generated for a peer with another public key")
            return False

        with self._recent_nonces.freeze():
            current_time = get_dht_time()
            if abs(auth.time - current_time) > self._MAX_CLIENT_SERVICER_TIME_DIFF.total_seconds():
                logger.debug("Clocks are not synchronized or a previous request is replayed again")
                return False
            if auth.nonce in self._recent_nonces:
                logger.debug("Previous request is replayed again")
                return False

        self._recent_nonces.store(
            auth.nonce, None, current_time + self._MAX_CLIENT_SERVICER_TIME_DIFF.total_seconds() * 3
        )
        return True

    async def sign_response(self, response: AuthorizedResponseBase, request: AuthorizedRequestBase) -> None:
        await self.refresh_token_if_needed()
        auth = response.auth

        auth.service_access_token.CopyFrom(self._local_access_token)
        auth.nonce = request.auth.nonce

        assert auth.signature == b""
        auth.signature = self._local_private_key.sign(response.SerializeToString())

    async def validate_response(self, response: AuthorizedResponseBase, request: AuthorizedRequestBase) -> bool:
        await self.refresh_token_if_needed()
        auth = response.auth

        if not self.is_token_valid(auth.service_access_token):
            logger.debug("Service failed to prove that it (still) has access to the network")
            return False

        service_public_key = RSAPublicKey.from_bytes(auth.service_access_token.public_key)
        signature = auth.signature
        auth.signature = b""
        if not service_public_key.verify(response.SerializeToString(), signature):
            logger.debug("Response has invalid signature")
            return False

        if auth.nonce != request.auth.nonce:
            logger.debug("Response is generated for another request")
            return False

        return True


class AuthRole(Enum):
    CLIENT = 0
    SERVICER = 1


class AuthRPCWrapper:
    def __init__(
        self,
        stub,
        role: AuthRole,
        authorizer: Optional[AuthorizerBase],
        service_public_key: Optional[RSAPublicKey] = None,
    ):
        self._stub = stub
        self._role = role
        self._authorizer = authorizer
        self._service_public_key = service_public_key

    def __getattribute__(self, name: str):
        if not name.startswith("rpc_"):
            return object.__getattribute__(self, name)

        method = getattr(self._stub, name)

        @functools.wraps(method)
        async def wrapped_rpc(request: AuthorizedRequestBase, *args, **kwargs):
            if self._authorizer is not None:
                if self._role == AuthRole.CLIENT:
                    await self._authorizer.sign_request(request, self._service_public_key)
                elif self._role == AuthRole.SERVICER:
                    if not await self._authorizer.validate_request(request):
                        return None

            response = await method(request, *args, **kwargs)

            if self._authorizer is not None:
                if self._role == AuthRole.SERVICER:
                    await self._authorizer.sign_response(response, request)
                elif self._role == AuthRole.CLIENT:
                    if not await self._authorizer.validate_response(response, request):
                        return None

            return response

        return wrapped_rpc



from dataclasses import dataclass
from typing import Tuple

from hivemind.p2p import PeerID


@dataclass(frozen=True)
class GroupInfo:
    """A group of peers assembled through decentralized matchmaking"""

    group_id: bytes  # random unique bytestring that describes the current group, generated by group leader
    peer_ids: Tuple[PeerID, ...]  # an ordered sequence of peer_ids of each groupmate
    gathered: Tuple[bytes, ...]  # binary metadata gathered from all peers by leader, same order as peer_ids

    @property
    def group_size(self):
        return len(self.peer_ids)

    def __contains__(self, peer_id: PeerID):
        return peer_id in self.peer_ids



import torch


class OptimizerWrapper(torch.optim.Optimizer):
    """A wrapper for pytorch.optim.Optimizer that forwards all methods to the wrapped optimizer"""

    def __init__(self, optim: torch.optim.Optimizer):
        super().__init__(optim.param_groups, optim.defaults)
        self.optim = optim

    @property
    def defaults(self):
        return self.optim.defaults

    @property
    def state(self):
        return self.optim.state

    def __getstate__(self):
        return self.optim.__getstate__()

    def __setstate__(self, state):
        self.optim.__setstate__(state)

    def __repr__(self):
        return f"{self.__class__.__name__}({repr(self.optim)})"

    def state_dict(self):
        return self.optim.state_dict()

    def load_state_dict(self, state_dict: dict) -> None:
        return self.optim.load_state_dict(state_dict)

    def step(self, *args, **kwargs):
        return self.optim.step(*args, **kwargs)

    def zero_grad(self, *args, **kwargs):
        return self.optim.zero_grad(*args, **kwargs)

    @property
    def param_groups(self):
        return self.optim.param_groups

    def add_param_group(self, param_group: dict) -> None:
        return self.optim.add_param_group(param_group)


class ClippingWrapper(OptimizerWrapper):
    """A wrapper of torch.Optimizer that clips gradients by global norm before each step"""

    def __init__(self, optim: torch.optim.Optimizer, clip_grad_norm: float):
        super().__init__(optim)
        self.clip_grad_norm = clip_grad_norm

    def step(self, *args, **kwargs):
        parameters = tuple(param for group in self.param_groups for param in group["params"])
        torch.nn.utils.clip_grad_norm_(parameters, self.clip_grad_norm)
        return super().step(*args, **kwargs)



import asyncio
import heapq
from collections import deque
from functools import partial
from typing import Deque, Dict, Iterator, List, Optional, Sequence, Set, Tuple, Union

from hivemind.dht import DHT, DHTNode
from hivemind.moe.client.expert import RemoteExpert, batch_create_remote_experts, create_remote_experts
from hivemind.moe.expert_uid import (
    FLAT_EXPERT,
    PREFIX_PATTERN,
    UID_DELIMITER,
    Coordinate,
    ExpertInfo,
    ExpertPrefix,
    ExpertUID,
    Score,
    is_valid_prefix,
    is_valid_uid,
)
from hivemind.p2p import PeerID
from hivemind.utils import DHTExpiration, MPFuture, ValueWithExpiration, get_dht_time, get_logger

logger = get_logger(__name__)


class MoEBeamSearcher:
    """
    Utility class that uses DHT to find most suitable experts for RemoteMixtureOfExperts.
    Each expert has an identifier in the form of {prefix}.{i}.{j}.{...}, e.g. "ffn_expert.98.76.54.32.10"
    An expert identifier consists of:

        * optional prefix that determines expert role, experiment name, etc.
        * one or more integers that determine that expert's position in an N-dimensional grid

    A hivemind.moe.Server can ``declare_experts(dht, expert_uids: List[str])`` to make its experts visible to everyone.
    When declaring experts, DHT will store each expert's uid and all its prefixes until :expiration: (specified at init)
    For instance, declaring "ffn_expert.98.76.54.32.10" will store the following keys in a DHT:
    ``"ffn_expert.98", "ffn_expert.98.76", "ffn_expert.98.76.54", ..., "ffn_expert.98.76.54.32.10"``

    In order to enable fast beam search, DHT maintains dictionaries of all active suffixes for every prefix
    (e.g. "ffn_expert.98": {76: ffn_expert.98.76...., 123: ffn_expert.98.123..., 225: ffn_expert.98.225....}))

    RemoteMixtureOfExperts can use these prefixes to find top-k most suitable experts with a left-to-right beam search.
    For instance, consider RemoteMixtureOfExperts with prefix "ffn_expert" and grid size [100, 100, 100, 100, 100].
    This MoE can query all experts with that prefix and arbitrary indices in 0...99 along each dimension.
    However, not every expert in such 100^5 grid can be alive at a given moment of time (the grid size is redundant).
    In order to find k best "alive" experts, MoE first ranks indices along the first dimension with its gating function.
    It can then check which of those indices correspond to "alive" experts by querying keys such as "ffn_expert.98".

    After selecting k best indices along first dimension, MoE moves to the second dimension.
    It can find top-k index pairs (e.g. "expert.98.76") that use one of k best indices from the previous step.
    This beam search explores one additional dimension per step and finds k best experts from across the DHT
    in O(k * num_dimensions * dimension_size) time depending on the chosen grid dimensions.

    :param dht: a running DHT daemon that is used for beam search AND local caching
    :param uid_prefix: search for experts whose uids start with this prefix
    :param grid_size: dimensions that form expert uid (see above)
    :param num_workers: number of concurrent DHT coroutines per beam search
    :param negative_caching: if True, whenever DHT is unable to find an expert or prefix, it will cache the "no key"
      result inside the DHT for :expiration: seconds. Caching only affects beam search and has three main effects:

      1. Faster beam search under node failures: if there are inconsistencies in DHT keys, such as a prefix pointing to
         a now-defunct expert, these inconsistencies will be overwritten by the first peer that stumbles upon them. As a
         result, beam search will not have to wait for non-existent experts until the expiration of their DHT entries;
      2. Delayed expert availability: Without negative cache, new experts are always immediately available for beam
         search after they are published to the DHT. With negative cache, there are rare cases (e.g. when adding new
         experts in place of recently defunct ones) when new experts will be initially invisible, but gradually become
         visible to more peers as those peers refresh their cache. This process takes at most :expiration: seconds;
      3. Faster beam search in very sparse grids: there is one edge case where negative cache will improve beam search
         performance; If an expert grid is very sparse, there can be empty indices in the first grid dimension (i.e.
         indices {i} such that _no_ experts that start with "{prefix}.{i}.*"). If so, the default beam search will
         be very slow due to the way it forms initial beam. Beam search with negative cache enabled will run normally.
         Though, this is a pathological case (e.g. only 90 experts in an oversized 100x100 grid) that should be avoided.
    """

    def __init__(
        self,
        dht: DHT,
        uid_prefix: ExpertPrefix,
        grid_size: Sequence[int],
        num_workers: Optional[int] = None,
        negative_caching: bool = True,
        cache_expiration: DHTExpiration = 300,
        **kwargs,
    ):
        if not uid_prefix.endswith(UID_DELIMITER):
            uid_prefix += UID_DELIMITER
            logger.info(f"Prefix must end with '{UID_DELIMITER}'. Changing to {uid_prefix}{UID_DELIMITER}")
        assert is_valid_prefix(uid_prefix), f"Prefix '{uid_prefix}' is invalid."
        self.dht = dht
        self.uid_prefix, self.grid_size = uid_prefix, grid_size
        self.total_grid_size = sum(grid_size)
        self.negative_caching, self.cache_expiration = negative_caching, cache_expiration
        self.num_workers, self.dht_kwargs = num_workers, kwargs

    def get_initial_beam(
        self, scores: Sequence[float], beam_size: int, return_future: bool = False
    ) -> List[Tuple[Score, ExpertPrefix, Dict[Coordinate, ExpertInfo]]]:
        """
        :param scores: prefer suffix coordinates that have highest scores
        :param beam_size: select this many active suffixes with highest scores
        :param return_future: if False (default), return when finished. Otherwise return MPFuture and run in background.
        :returns: a list of up to beam_size tuples of (prefix score, prefix itself, dict{suffix: example expert})
        """
        return self.dht.run_coroutine(
            partial(
                self._get_initial_beam,
                prefix=self.uid_prefix,
                beam_size=beam_size,
                scores=tuple(scores),
                negative_caching=self.negative_caching,
                cache_expiration=self.cache_expiration,
                num_workers=self.num_workers,
            ),
            return_future,
        )

    @staticmethod
    async def _get_initial_beam(
        dht: DHT,
        node: DHTNode,
        prefix: ExpertPrefix,
        beam_size: int,
        scores: Tuple[float, ...],
        negative_caching: bool,
        cache_expiration: DHTExpiration,
        num_workers: Optional[int] = None,
    ) -> List[Tuple[Score, ExpertPrefix, Dict[Coordinate, ExpertInfo]]]:
        num_workers = num_workers or dht.num_workers or beam_size
        beam: List[Tuple[Score, ExpertPrefix, Dict[Coordinate, ExpertInfo]]] = []
        unattempted_indices: List[Coordinate] = sorted(
            range(len(scores)), key=scores.__getitem__
        )  # from worst to best
        pending_tasks: Deque[Tuple[Coordinate, ExpertPrefix, asyncio.Task]] = deque()

        while len(beam) < beam_size and (unattempted_indices or pending_tasks):
            # dispatch additional tasks
            while unattempted_indices and len(pending_tasks) < num_workers:
                next_index = unattempted_indices.pop()  # note: this is best unattempted index because of sort order
                next_best_prefix = f"{prefix}{next_index}{UID_DELIMITER}"
                pending_tasks.append((next_index, next_best_prefix, asyncio.create_task(node.get(next_best_prefix))))

            # await the next best prefix to be fetched
            pending_best_index, pending_best_prefix, pending_task = pending_tasks.popleft()
            try:
                maybe_prefix_data = await pending_task
                if maybe_prefix_data is not None and isinstance(maybe_prefix_data.value, dict):
                    successors = MoEBeamSearcher._select_valid_entries(maybe_prefix_data)
                    if successors:
                        beam.append((scores[pending_best_index], pending_best_prefix, successors))
                elif maybe_prefix_data is None and negative_caching:
                    logger.debug(f"DHT negative caching: storing a 'no prefix' entry for {pending_best_prefix}")
                    asyncio.create_task(
                        node.store(
                            pending_best_prefix,
                            subkey=-1,
                            value=None,
                            expiration_time=get_dht_time() + cache_expiration,
                        )
                    )

            except asyncio.CancelledError:
                for _, pending_task in pending_tasks:
                    pending_task.cancel()
                raise
        return beam

    def get_active_successors(
        self, prefixes: List[ExpertPrefix], grid_size: Optional[int] = None, return_future: bool = False
    ) -> Dict[ExpertPrefix, Dict[Coordinate, ExpertInfo]]:
        """
        :param prefixes: a list of prefix for which to find active successor uids
        :param grid_size: if specified, only return successors if they are in range [0, grid_size)
        :param return_future: if False (default), find and return successors. Otherwise return MPFuture and fill later.
        :returns: for every expert, return a dict{active_next_coordinate: (matching_expert_uid, matching_endpoint)}
        :note: if a prefix is not found, get_active_successors will return an empty dictionary for that prefix
        """
        assert not isinstance(prefixes, str), "Please send a list / tuple of expert prefixes."
        for prefix in prefixes:
            assert is_valid_prefix(prefix), f"prefix '{prefix}' is invalid, it must follow {PREFIX_PATTERN.pattern}"
        return self.dht.run_coroutine(
            partial(
                self._get_active_successors,
                prefixes=list(prefixes),
                grid_size=grid_size,
                negative_caching=self.negative_caching,
                cache_expiration=self.cache_expiration,
                num_workers=self.num_workers,
            ),
            return_future=return_future,
        )

    @staticmethod
    def _select_valid_entries(entry: ValueWithExpiration, grid_size: Optional[int] = None):
        if not isinstance(entry, ValueWithExpiration) or not isinstance(entry.value, dict):
            return {}
        return {
            coord: ExpertInfo(uid=match.value[0], peer_id=PeerID.from_base58(match.value[1]))
            for coord, match in entry.value.items()
            if isinstance(coord, Coordinate)
            and (grid_size is None or 0 <= coord < grid_size)
            and isinstance(match, ValueWithExpiration)
            and isinstance(match.value, tuple)
            and len(match.value) == 2
            and is_valid_uid(match.value[0])
            and isinstance(match.value[1], str)
        }

    @staticmethod
    async def _get_active_successors(
        dht: DHT,
        node: DHTNode,
        prefixes: List[ExpertPrefix],
        grid_size: Optional[int],
        negative_caching: bool,
        cache_expiration: DHTExpiration,
        num_workers: Optional[int] = None,
    ) -> Dict[ExpertPrefix, Dict[Coordinate, ExpertInfo]]:
        grid_size = grid_size or float("inf")
        num_workers = num_workers or min(len(prefixes), dht.num_workers or len(prefixes))
        dht_responses = await node.get_many(keys=prefixes, num_workers=num_workers)
        successors: Dict[ExpertPrefix, Dict[Coordinate, ExpertInfo]] = {}
        for prefix, found in dht_responses.items():
            successors[prefix] = MoEBeamSearcher._select_valid_entries(found, grid_size)
            if not successors[prefix] and negative_caching:
                logger.debug(f"DHT negative caching: storing a 'no prefix' entry for {prefix}")
                asyncio.create_task(
                    node.store(prefix, subkey=-1, value=None, expiration_time=get_dht_time() + cache_expiration)
                )
        return successors

    def find_best_experts(
        self, grid_scores: Sequence[Sequence[float]], beam_size: int, return_future: bool = False
    ) -> Union[List[RemoteExpert], MPFuture[List[RemoteExpert]]]:
        """
        Find and return :beam_size: active experts with highest scores, use both local cache and DHT

        :param grid_scores: scores predicted for each dimension in the grid
        :type grid_scores: model scores for each grid dimension, list of arrays of shape grid_size[i]
        :param beam_size: how many best experts should beam search return
         After time_budget is reached, beam search won't search for more experts and instead fall back on local cache
         Please note that any queries that fall outside the budget will still be performed in background and cached
         for subsequent iterations as long as DHTNode.cache_locally is True
        :param return_future: if set to True, returns MPFuture that can be awaited to get the actual result
        :returns: a list that contains *up to* k_best RemoteExpert instances
        """
        assert len(grid_scores) == len(self.grid_size) and beam_size > 0
        result = self.dht.run_coroutine(
            partial(
                self._find_best_experts,
                prefix=self.uid_prefix,
                beam_size=beam_size,
                grid_scores=list(grid_scores),
                negative_caching=self.negative_caching,
                cache_expiration=self.cache_expiration,
                num_workers=self.num_workers,
            ),
            return_future,
        )
        return create_remote_experts(result, self.dht, return_future)

    @classmethod
    async def _find_best_experts(
        cls,
        dht: DHT,
        node: DHTNode,
        prefix: str,
        grid_scores: List[Tuple[float]],
        beam_size: int,
        negative_caching: bool,
        cache_expiration: DHTExpiration,
        num_workers: Optional[int] = None,
    ) -> List[ExpertInfo]:
        num_workers = num_workers or min(beam_size, dht.num_workers or beam_size)

        # form initial beam from top-k active L1 prefixes, each row is (score, uid prefix, possible suffixes)
        beam: List[Tuple[Score, ExpertPrefix, Dict[Coordinate, ExpertInfo]]] = await cls._get_initial_beam(
            dht, node, prefix, beam_size, grid_scores[0], negative_caching, min(beam_size, num_workers)
        )

        best_experts_heap: List[Tuple[Score, ExpertInfo]] = []  # max-heap of expert infos ordered by scores
        unique_experts: Set[ExpertUID] = set()

        for dim_index in range(1, len(grid_scores) - 1):
            for score, expert_info in cls._iterate_matching_experts(beam, grid_scores):
                if expert_info.uid not in unique_experts:
                    push_and_maybe_pop = heapq.heappush if len(best_experts_heap) < beam_size else heapq.heappushpop
                    push_and_maybe_pop(best_experts_heap, (score, expert_info))
                    unique_experts.add(expert_info.uid)

            # form new beam using successors from the current beam
            dim_scores = grid_scores[dim_index]
            best_active_pairs: List[Tuple[Score, ExpertPrefix]] = heapq.nlargest(
                beam_size,
                (
                    (prefix_score + dim_scores[next_coord], f"{prefix}{next_coord}{UID_DELIMITER}")
                    for prefix_score, prefix, suffixes in beam
                    for next_coord in suffixes.keys()
                    if isinstance(next_coord, int) and 0 <= next_coord < len(dim_scores)
                ),
            )

            _, best_uid_prefixes = zip(*best_active_pairs)

            # search DHT for next step suffixes
            successors = await cls._get_active_successors(
                dht,
                node,
                best_uid_prefixes,
                grid_size=None,
                negative_caching=negative_caching,
                cache_expiration=cache_expiration,
                num_workers=num_workers,
            )
            beam = [(score, prefix, successors[prefix]) for score, prefix in best_active_pairs if successors[prefix]]
            if not beam:
                logger.warning(f"Beam search had to terminate prematurely because of empty beam (dim 0)")
                break

        # add best experts from the final beam
        for score, expert_info in cls._iterate_matching_experts(beam, grid_scores):
            if expert_info.uid not in unique_experts:
                push_and_maybe_pop = heapq.heappush if len(best_experts_heap) < beam_size else heapq.heappushpop
                push_and_maybe_pop(best_experts_heap, (score, expert_info))
                unique_experts.add(expert_info.uid)

        return [expert_info for _, expert_info in sorted(best_experts_heap, reverse=True)]

    @staticmethod
    def _iterate_matching_experts(
        beam: List[Tuple[Score, ExpertPrefix, Dict[Coordinate, ExpertInfo]]], grid_scores: Sequence[Sequence[float]]
    ) -> Iterator[Tuple[Score, ExpertInfo]]:
        """iterate over all exemplar experts attached to current beam"""
        for score, prefix, suffixes in beam:
            for next_coord, match in suffixes.items():
                if len(grid_scores) == 1 and next_coord == FLAT_EXPERT:
                    yield score, match
                elif isinstance(match.uid, ExpertUID) and match.uid.count(UID_DELIMITER) == len(grid_scores):
                    expert_coords = match.uid.split(UID_DELIMITER)[1:]
                    if all(
                        coord.isdigit() and 0 <= int(coord) < len(grid_scores[i])
                        for i, coord in enumerate(expert_coords)
                    ):
                        expert_score = sum(
                            scores[coord] for scores, coord in zip(grid_scores, map(int, expert_coords))
                        )
                        yield expert_score, match
                    else:
                        logger.warning(f"Found incompatible expert coordinates: {expert_coords}")
                else:
                    logger.warning(f"Found incompatible expert UID: {match.uid}")

    def batch_find_best_experts(
        self, batch_grid_scores: Sequence[Sequence[Sequence[float]]], beam_size: int, return_future: bool = False
    ) -> Union[List[List[RemoteExpert]], MPFuture[List[List[RemoteExpert]]]]:
        """
        Find and return :beam_size: active experts with highest scores, use both local cache and DHT

        :param batch_grid_scores: scores predicted for each batch example and each dimension in the grid,
        :type batch_grid_scores: list of arrays of shape (batch_size, grid_size[i])
        :param beam_size: how many best experts should beam search return
         After time_budget is reached, beam search won't search for more experts and instead fall back on local cache
         Please note that any queries that fall outside the budget will still be performed in background and cached
         for subsequent iterations as long as DHTNode.cache_locally is True
        :param return_future: if set to True, returns MPFuture that can be awaited to get the actual result
        :returns: a list that contains *up to* k_best RemoteExpert instances
        """
        result = self.dht.run_coroutine(
            partial(
                self._batch_find_best_experts,
                prefix=self.uid_prefix,
                batch_grid_scores=batch_grid_scores,
                beam_size=beam_size,
                negative_caching=self.negative_caching,
                num_workers=self.num_workers,
            ),
            return_future,
        )

        return batch_create_remote_experts(result, self.dht, return_future)

    @classmethod
    async def _batch_find_best_experts(
        cls,
        dht: DHT,
        node: DHTNode,
        prefix: str,
        batch_grid_scores: Sequence[Sequence[Tuple[float]]],
        beam_size: int,
        negative_caching: bool,
        num_workers: Optional[int],
    ) -> Sequence[Sequence[ExpertInfo]]:
        batch_grid_scores = [
            [tuple(grid_score[i]) for grid_score in batch_grid_scores] for i in range(len(batch_grid_scores[0]))
        ]
        coros = [
            cls._find_best_experts(dht, node, prefix, grid_scores, beam_size, negative_caching, num_workers)
            for grid_scores in batch_grid_scores
        ]
        return await asyncio.gather(*coros)



from hivemind.utils.logging import get_logger

logger = get_logger(__name__)


def increase_file_limit(new_soft=2**15, new_hard=2**15):
    """Increase the maximum number of open files. On Linux, this allows spawning more processes/threads."""
    try:
        import resource  # local import to avoid ImportError for Windows users

        soft, hard = resource.getrlimit(resource.RLIMIT_NOFILE)
        new_soft = max(soft, new_soft)
        new_hard = max(hard, new_hard)
        logger.info(f"Increasing file limit: soft {soft}=>{new_soft}, hard {hard}=>{new_hard}")
        return resource.setrlimit(resource.RLIMIT_NOFILE, (new_soft, new_hard))
    except Exception as e:
        logger.warning(f"Failed to increase file limit: {e}")



import asyncio
import inspect
from dataclasses import dataclass
from typing import Any, AsyncIterator, List, Optional, Tuple, Type, get_type_hints

from hivemind.p2p.p2p_daemon import P2P
from hivemind.p2p.p2p_daemon_bindings.datastructures import PeerID


@dataclass
class RPCHandler:
    method_name: str
    request_type: type
    response_type: type
    stream_input: bool
    stream_output: bool


class StubBase:
    """
    Base class for P2P RPC stubs. The interface mimicks gRPC stubs.

    Servicer derives stub classes for particular services (e.g. DHT, averager, etc.) from StubBase,
    adding the necessary rpc_* methods. Calls to these methods are translated to calls to the remote peer.
    """

    def __init__(self, p2p: P2P, peer: PeerID, namespace: Optional[str]):
        self._p2p = p2p
        self._peer = peer
        self._namespace = namespace


class ServicerBase:
    """
    Base class for P2P RPC servicers (e.g. DHT, averager, MoE server). The interface mimicks gRPC servicers.

    - ``add_p2p_handlers(self, p2p)`` registers all rpc_* methods of the derived class as P2P handlers, allowing
      other peers to call them. It uses type annotations for the ``request`` parameter and the return value
      to infer protobufs the methods operate with.

    - ``get_stub(self, p2p, peer)`` creates a stub with all rpc_* methods. Calls to the stub methods are translated
      to calls to the remote peer.
    """

    _rpc_handlers: Optional[List[RPCHandler]] = None
    _stub_type: Optional[Type[StubBase]] = None

    @classmethod
    def _collect_rpc_handlers(cls) -> None:
        if cls._rpc_handlers is not None:
            return

        cls._rpc_handlers = []
        for method_name, method in inspect.getmembers(cls, predicate=inspect.isfunction):
            if method_name.startswith("rpc_"):
                spec = inspect.getfullargspec(method)
                if len(spec.args) < 3:
                    raise ValueError(
                        f"{method_name} is expected to at least three positional arguments "
                        f"(self, request: TInputProtobuf | AsyncIterator[TInputProtobuf], context: P2PContext)"
                    )
                request_arg = spec.args[1]
                hints = get_type_hints(method)
                try:
                    request_type = hints[request_arg]
                    response_type = hints["return"]
                except KeyError:
                    raise ValueError(
                        f"{method_name} is expected to have type annotations "
                        f"like `dht_pb2.FindRequest` or `AsyncIterator[dht_pb2.FindRequest]` "
                        f"for the `{request_arg}` parameter and the return value"
                    )
                request_type, stream_input = cls._strip_iterator_hint(request_type)
                response_type, stream_output = cls._strip_iterator_hint(response_type)

                cls._rpc_handlers.append(
                    RPCHandler(method_name, request_type, response_type, stream_input, stream_output)
                )

        cls._stub_type = type(
            f"{cls.__name__}Stub",
            (StubBase,),
            {handler.method_name: cls._make_rpc_caller(handler) for handler in cls._rpc_handlers},
        )

    @classmethod
    def _make_rpc_caller(cls, handler: RPCHandler):
        input_type = AsyncIterator[handler.request_type] if handler.stream_input else handler.request_type
        output_type = AsyncIterator[handler.response_type] if handler.stream_output else handler.response_type

        # This method will be added to a new Stub type (a subclass of StubBase)
        async def caller(self: StubBase, input: input_type, timeout: Optional[float] = None) -> output_type:
            handle_name = cls._get_handle_name(self._namespace, handler.method_name)
            if not handler.stream_output:
                return await asyncio.wait_for(
                    self._p2p.call_protobuf_handler(self._peer, handle_name, input, handler.response_type),
                    timeout=timeout,
                )

            if timeout is not None:
                raise ValueError("Timeouts for handlers returning streams are not supported")
            return await self._p2p.iterate_protobuf_handler(self._peer, handle_name, input, handler.response_type)

        caller.__name__ = handler.method_name
        return caller

    async def add_p2p_handlers(
        self, p2p: P2P, wrapper: Any = None, *, namespace: Optional[str] = None, balanced: bool = False
    ) -> None:
        self._collect_rpc_handlers()

        servicer = self if wrapper is None else wrapper
        await asyncio.gather(
            *[
                p2p.add_protobuf_handler(
                    self._get_handle_name(namespace, handler.method_name),
                    getattr(servicer, handler.method_name),
                    handler.request_type,
                    stream_input=handler.stream_input,
                    stream_output=handler.stream_output,
                    balanced=balanced,
                )
                for handler in self._rpc_handlers
            ]
        )

    async def remove_p2p_handlers(self, p2p: P2P, *, namespace: Optional[str] = None) -> None:
        self._collect_rpc_handlers()

        await asyncio.gather(
            *[
                p2p.remove_protobuf_handler(
                    self._get_handle_name(namespace, handler.method_name),
                    stream_input=handler.stream_input,
                    stream_output=handler.stream_output,
                )
                for handler in self._rpc_handlers
            ]
        )

    @classmethod
    def get_stub(cls, p2p: P2P, peer: PeerID, *, namespace: Optional[str] = None) -> StubBase:
        cls._collect_rpc_handlers()
        return cls._stub_type(p2p, peer, namespace)

    @classmethod
    def _get_handle_name(cls, namespace: Optional[str], method_name: str) -> str:
        handle_name = f"{cls.__name__}.{method_name}"
        if namespace is not None:
            handle_name = f"{namespace}::{handle_name}"
        return handle_name

    @staticmethod
    def _strip_iterator_hint(hint: type) -> Tuple[type, bool]:
        if hasattr(hint, "_name") and hint._name in ("AsyncIterator", "AsyncIterable"):
            return hint.__args__[0], True

        return hint, False



import dataclasses
import re
from typing import Optional

from hivemind.dht.validation import DHTRecord, RecordValidatorBase
from hivemind.utils import MSGPackSerializer, get_logger
from hivemind.utils.crypto import RSAPrivateKey, RSAPublicKey

logger = get_logger(__name__)


class RSASignatureValidator(RecordValidatorBase):
    """
    Introduces a notion of *protected records* whose key/subkey contains substring
    "[owner:ssh-rsa ...]" with an RSA public key of the owner.

    If this validator is used, changes to such records always must be signed with
    the corresponding private key (so only the owner can change them).
    """

    PUBLIC_KEY_FORMAT = b"[owner:_key_]"
    SIGNATURE_FORMAT = b"[signature:_value_]"

    PUBLIC_KEY_REGEX = re.escape(PUBLIC_KEY_FORMAT).replace(b"_key_", rb"(.+?)")
    _PUBLIC_KEY_RE = re.compile(PUBLIC_KEY_REGEX)
    _SIGNATURE_RE = re.compile(re.escape(SIGNATURE_FORMAT).replace(b"_value_", rb"(.+?)"))

    _cached_private_key = None

    def __init__(self, private_key: Optional[RSAPrivateKey] = None):
        if private_key is None:
            private_key = RSAPrivateKey.process_wide()
        self._private_key = private_key

        serialized_public_key = private_key.get_public_key().to_bytes()
        self._local_public_key = self.PUBLIC_KEY_FORMAT.replace(b"_key_", serialized_public_key)

    @property
    def local_public_key(self) -> bytes:
        return self._local_public_key

    def validate(self, record: DHTRecord) -> bool:
        public_keys = self._PUBLIC_KEY_RE.findall(record.key)
        if record.subkey is not None:
            public_keys += self._PUBLIC_KEY_RE.findall(record.subkey)
        if not public_keys:
            return True  # The record is not protected with a public key

        if len(set(public_keys)) > 1:
            logger.debug(f"Key and subkey can't contain different public keys in {record}")
            return False
        public_key = RSAPublicKey.from_bytes(public_keys[0])

        signatures = self._SIGNATURE_RE.findall(record.value)
        if len(signatures) != 1:
            logger.debug(f"Record should have exactly one signature in {record}")
            return False
        signature = signatures[0]

        stripped_record = dataclasses.replace(record, value=self.strip_value(record))
        if not public_key.verify(self._serialize_record(stripped_record), signature):
            logger.debug(f"Signature is invalid in {record}")
            return False
        return True

    def sign_value(self, record: DHTRecord) -> bytes:
        if self._local_public_key not in record.key and self._local_public_key not in record.subkey:
            return record.value

        signature = self._private_key.sign(self._serialize_record(record))
        return record.value + self.SIGNATURE_FORMAT.replace(b"_value_", signature)

    def strip_value(self, record: DHTRecord) -> bytes:
        return self._SIGNATURE_RE.sub(b"", record.value)

    def _serialize_record(self, record: DHTRecord) -> bytes:
        return MSGPackSerializer.dumps(dataclasses.astuple(record))

    @property
    def priority(self) -> int:
        # On validation, this validator must be executed before validators
        # that deserialize the record
        return 10

    def merge_with(self, other: RecordValidatorBase) -> bool:
        if not isinstance(other, RSASignatureValidator):
            return False

        # Ignore another RSASignatureValidator instance (it doesn't make sense to have several
        # instances of this class) and report successful merge
        return True



"""
Compression strategies that reduce the network communication in .averaging, .optim and .moe
"""

from hivemind.compression.adaptive import PerTensorCompression, RoleAdaptiveCompression, SizeAdaptiveCompression
from hivemind.compression.base import CompressionBase, CompressionInfo, NoCompression, TensorRole
from hivemind.compression.floating import Float16Compression, ScaledFloat16Compression
from hivemind.compression.quantization import BlockwiseQuantization, Quantile8BitQuantization, Uniform8BitQuantization
from hivemind.compression.serialization import (
    deserialize_tensor_stream,
    deserialize_torch_tensor,
    serialize_torch_tensor,
)



name_to_block = {}
name_to_input = {}

import hivemind.moe.server.layers.common
import hivemind.moe.server.layers.dropout
from hivemind.moe.server.layers.custom_experts import add_custom_models_from_file, register_expert_class
from hivemind.moe.server.layers.lr_schedule import get_linear_schedule_with_warmup

schedule_name_to_scheduler = {"linear": get_linear_schedule_with_warmup, "none": None}



from typing import Optional, Sequence, Tuple

import numpy as np
import scipy.optimize

from hivemind.utils.logging import get_logger

logger = get_logger(__name__)

LOAD_BALANCING_LP_DECIMALS = 9


def load_balance_peers(vector_size, bandwidths: Sequence[Optional[float]], min_size: int = 0) -> Tuple[int, ...]:
    """
    Find an optimal partitioning of weights for butterfly all-reduce given peer bandwidths.
    :param vector_size: total size of the averaged vector (in elements, not bytes)
    :param bandwidths: 1d array of non-negative bandwidths for each peer capable of averaging
      zeros stand for client-only participants, None represents "not specified" (resolved as mean of other pears)
    :param min_size: peers that can aggregate less than this many elements will be assigned nothing
    :returns: an integer array where i-th element is the number of weights assigned to i-th peer
    """
    specified_bandwidth = [item for item in bandwidths if item is not None and item > 0]

    if specified_bandwidth:
        default_bandwidth = np.mean(specified_bandwidth)
        bandwidths = [item if item is not None else default_bandwidth for item in bandwidths]
        scores = optimize_parts_lp(vector_size, np.asarray(bandwidths), min_size)
    else:
        assert not all(item == 0 for item in bandwidths), "Must have at least one nonzero bandwidth"
        scores = np.asarray([1.0 if item is None else 0.0 for item in bandwidths])

    # TODO(jheuristic) we no longer need hagenbach-bishoff with new AllReduceRunner
    return tuple(hagenbach_bishoff(vector_size, scores))


def optimize_parts_lp(vector_size: int, bandwidths: np.ndarray, min_size: int = 0) -> np.ndarray:
    """
    This method solves an optimization problem to minimize the total allreduce time.
    In butterfly all-reduce, each peer acts both as a "client" and as an "aggregator":
    * a "client" splits his local vector into shards and sends each shard to one peer, then downloads the average
    * an "aggregator" receives a certain part of vector components from all peers, aggregates and returns the average

    Peer i network load as a "client" = vector_size * (1 - fraction_assigned_to_peer_i)
    Peer i network load as an "aggregator" = vector_size * (group_size - 1) * fraction_assigned_to_peer_i
    Peer i total communication = vector_size * [1 + (group_size - 2) * fraction_assigned_to_peer_i]
    Total time = max_i (total_communication_for_peer_i / bandwidths[i])

    We solve this optimization problem by reducing it to linear programming with a minimax reduction
    (see lecture notes: https://www.usna.edu/Users/math/dphillip/sa305.s15/phillips/lessons/32/32.pdf )

    :returns: a vector of "scores", i-th score is proportional to the fraction of weights assigned to i-th peer
    """
    assert np.all(bandwidths >= 0) and np.any(bandwidths > 0)
    bandwidths = np.asarray(bandwidths, dtype=np.float64)
    permutation = np.argsort(-bandwidths)
    bandwidths = bandwidths[permutation]
    is_nonzero = bandwidths != 0

    group_size = len(bandwidths)
    num_variables = group_size + 1  # [w_1, ..., w_N, xi]

    c = np.zeros(num_variables, dtype=np.float64)
    c[-1] = 1.0  # optimize w.r.t. xi

    # the constraints below are tuples (A, b) such that Ax <= b
    nonnegative_weights = -np.eye(group_size, num_variables, dtype=c.dtype), np.zeros(group_size, c.dtype)
    weights_sum_to_one = c[None, :] - 1.0, np.array([-1.0])
    coeff_per_variable = (group_size - 2.0) / np.maximum(bandwidths, 10**-LOAD_BALANCING_LP_DECIMALS)
    coeff_matrix_minus_xi = np.hstack([np.diag(coeff_per_variable), -np.ones((group_size, 1), c.dtype)])
    xi_is_maximum = coeff_matrix_minus_xi[is_nonzero], -1.0 / bandwidths[is_nonzero]
    force_max_weights = np.eye(group_size, M=num_variables, dtype=c.dtype), is_nonzero.astype(c.dtype)

    A, b = list(map(np.concatenate, zip(nonnegative_weights, weights_sum_to_one, xi_is_maximum, force_max_weights)))

    solution = scipy.optimize.linprog(c, A_ub=A, b_ub=b, method="interior-point")
    if solution.success:
        peer_scores = solution.x[:group_size]
        # if some peers have less than min_size elements, transfer their share to other peers (if any)
        if np.max(peer_scores) >= min_size / float(vector_size):
            peer_scores[peer_scores < min_size / float(vector_size)] = 0.0
        peer_scores = np.round(peer_scores, LOAD_BALANCING_LP_DECIMALS)
    else:
        logger.error(f"Failed to solve load-balancing for bandwidths {bandwidths}")
        peer_scores = np.ones(group_size, c.dtype)

    return peer_scores[np.argsort(permutation)]


def hagenbach_bishoff(vector_size: int, scores: Sequence[float]) -> Sequence[int]:
    """
    Split a vector between participants based on continuous fractions.
    https://en.wikipedia.org/wiki/Hagenbach-Bischoff_system
    The code is based on https://github.com/crflynn/voting

    :param vector_size: the total number of elements to be split
    :param scores: real-valued vector fractions for each peer
    :returns: integer-valued partitions assigned to every peer
    """
    total_score = sum(scores)
    allocated = [int(vector_size * score_i / total_score) for score_i in scores]
    while sum(allocated) < vector_size:
        quotients = [score / (allocated[idx] + 1) for idx, score in enumerate(scores)]
        idx_max = quotients.index(max(quotients))
        allocated[idx_max] += 1
    return allocated



import dataclasses
from abc import ABC, abstractmethod
from typing import Iterable


@dataclasses.dataclass(init=True, repr=True, frozen=True)
class DHTRecord:
    key: bytes
    subkey: bytes
    value: bytes
    expiration_time: float


class RecordValidatorBase(ABC):
    """
    Record validators are a generic mechanism for checking the DHT records including:
      - Enforcing a data schema (e.g. checking content types)
      - Enforcing security requirements (e.g. allowing only the owner to update the record)
    """

    @abstractmethod
    def validate(self, record: DHTRecord) -> bool:
        """
        Should return whether the `record` is valid.
        The valid records should have been extended with sign_value().

        validate() is called when another DHT peer:
          - Asks us to store the record
          - Returns the record by our request
        """

        pass

    def sign_value(self, record: DHTRecord) -> bytes:
        """
        Should return `record.value` extended with the record's signature.

        Note: there's no need to overwrite this method if a validator doesn't use a signature.

        sign_value() is called after the application asks the DHT to store the record.
        """

        return record.value

    def strip_value(self, record: DHTRecord) -> bytes:
        """
        Should return `record.value` stripped of the record's signature.
        strip_value() is only called if validate() was successful.

        Note: there's no need to overwrite this method if a validator doesn't use a signature.

        strip_value() is called before the DHT returns the record by the application's request.
        """

        return record.value

    @property
    def priority(self) -> int:
        """
        Defines the order of applying this validator with respect to other validators.

        The validators are applied:
          - In order of increasing priority for signing a record
          - In order of decreasing priority for validating and stripping a record
        """

        return 0

    def merge_with(self, other: "RecordValidatorBase") -> bool:
        """
        By default, all validators are applied sequentially (i.e. we require all validate() calls
        to return True for a record to be validated successfully).

        However, you may want to define another policy for combining your validator classes
        (e.g. for schema validators, we want to require only one validate() call to return True
        because each validator bears a part of the schema).

        This can be achieved with overriding merge_with(). It should:

          - Return True if it has successfully merged the `other` validator to `self`,
            so that `self` became a validator that combines the old `self` and `other` using
            the necessary policy. In this case, `other` should remain unchanged.

          - Return False if the merging has not happened. In this case, both `self` and `other`
            should remain unchanged. The DHT will try merging `other` to another validator or
            add it as a separate validator (to be applied sequentially).
        """

        return False


class CompositeValidator(RecordValidatorBase):
    def __init__(self, validators: Iterable[RecordValidatorBase] = ()):
        self._validators = []
        self.extend(validators)

    def extend(self, validators: Iterable[RecordValidatorBase]) -> None:
        for new_validator in validators:
            for existing_validator in self._validators:
                if existing_validator.merge_with(new_validator):
                    break
            else:
                self._validators.append(new_validator)
        self._validators.sort(key=lambda item: item.priority)

    def validate(self, record: DHTRecord) -> bool:
        for i, validator in enumerate(reversed(self._validators)):
            if not validator.validate(record):
                return False
            if i < len(self._validators) - 1:
                record = dataclasses.replace(record, value=validator.strip_value(record))
        return True

    def sign_value(self, record: DHTRecord) -> bytes:
        for validator in self._validators:
            record = dataclasses.replace(record, value=validator.sign_value(record))
        return record.value

    def strip_value(self, record: DHTRecord) -> bytes:
        for validator in reversed(self._validators):
            record = dataclasses.replace(record, value=validator.strip_value(record))
        return record.value



import importlib
import os
from typing import Callable, Type

import torch
import torch.nn as nn

from hivemind.moe.server.layers import name_to_block, name_to_input


def add_custom_models_from_file(path: str):
    spec = importlib.util.spec_from_file_location("custom_module", os.path.abspath(path))
    foo = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(foo)


def register_expert_class(name: str, sample_input: Callable[[int, int], torch.tensor]):
    """
    Adds a custom user expert to hivemind server.
    :param name: the name of the expert. It shouldn't coincide with existing modules\
        ('ffn', 'transformer', 'nop', 'det_dropout')
    :param sample_input: a function which gets batch_size and hid_dim and outputs a \
        sample of an input in the module
    :unchanged module
    """

    def _register_expert_class(custom_class: Type[nn.Module]):
        if name in name_to_block or name in name_to_input:
            raise RuntimeError("The class might already exist or be added twice")
        name_to_block[name] = custom_class
        name_to_input[name] = sample_input

        return custom_class

    return _register_expert_class



from hivemind.p2p.p2p_daemon import P2P, P2PContext
from hivemind.p2p.p2p_daemon_bindings import P2PDaemonError, P2PHandlerError, PeerID, PeerInfo
from hivemind.p2p.servicer import ServicerBase, StubBase
